import collections
import copy
import typing

from creme import stats
from creme import optim
from creme import utils

from . import base


__all__ = ['Baseline']


class Baseline(base.Recommender):
    """Baseline for recommender systems.

    A first-order approximation of the bias involved in target. The model equation is defined as:

    $$\\hat{y}(x) = \\bar{y} + bu_{u} + bi_{i}$$

    Where :math:`bu_{u}` and :math:`bi_{i}` are respectively the user and item biases. The model
    expect dict inputs containing both a `user` and an `item` entries.

    Parameters:
        optimizer: The sequential optimizer used for updating the weights.
        loss: The loss function to optimize for.
        l2: regularization amount used to push weights towards 0.
        initializer: Weights initialization scheme.
        clip_gradient: Clips the absolute value of each gradient value.

    Attributes:
        global_mean (stats.Mean): The target arithmetic mean.
        u_biases (collections.defaultdict): The user bias weights.
        i_biases (collections.defaultdict): The item bias weights.
        u_optimizer (optim.Optimizer): The sequential optimizer used for updating the user bias
            weights.
        i_optimizer (optim.Optimizer): The sequential optimizer used for updating the item bias
            weights.

    Example:

        >>> from creme import optim
        >>> from creme import reco

        >>> X_y = (
        ...     ({'user': 'Alice', 'item': 'Superman'}, 8),
        ...     ({'user': 'Alice', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Alice', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Alice', 'item': 'Notting Hill'}, 2),
        ...     ({'user': 'Alice', 'item': 'Harry Potter'}, 5),
        ...     ({'user': 'Bob', 'item': 'Superman'}, 8),
        ...     ({'user': 'Bob', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Bob', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Bob', 'item': 'Notting Hill'}, 2)
        ... )

        >>> model = reco.Baseline(optimizer=optim.SGD(0.005))

        >>> for x, y in X_y:
        ...     _ = model.fit_one(x, y)

        >>> model.predict_one({'user': 'Bob', 'item': 'Harry Potter'})
        6.538120

    Note:
        This model expects a dict input with a `user` and an `item` entries without any
        type constraint on their values (i.e. can be strings or numbers). Other entries are ignored.

    References:
        1. [Matrix factorization techniques for recommender systems](https://datajobs.com/data-science-repo/Recommender-Systems-[Netflix].pdf)

    """

    def __init__(self, optimizer: optim.Optimizer = None, loss: optim.losses.Loss = None,
                 l2=0., initializer: optim.initializers.Initializer = None, clip_gradient=1e12):
        self.optimizer = optim.SGD() if optimizer is None else copy.deepcopy(optimizer)
        self.u_optimizer = optim.SGD() if optimizer is None else copy.deepcopy(optimizer)
        self.i_optimizer = optim.SGD() if optimizer is None else copy.deepcopy(optimizer)
        self.loss = optim.losses.Squared() if loss is None else loss
        self.l2 = l2

        if initializer is None:
            initializer = optim.initializers.Zeros()
        self.initializer = initializer

        self.clip_gradient = clip_gradient
        self.global_mean = stats.Mean()
        self.u_biases: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(initializer)
        self.i_biases: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(initializer)

    def _predict_one(self, user, item):
        return self.global_mean.get() + self.u_biases[user] + self.i_biases[item]

    def _fit_one(self, user, item, y):

        # Update the global mean
        self.global_mean.update(y)

        # Calculate the gradient of the loss with respect to the prediction
        g_loss = self.loss.gradient(y, self._predict_one(user, item))

        # Clamp the gradient to avoid numerical instability
        g_loss = utils.math.clamp(g_loss, minimum=-self.clip_gradient, maximum=self.clip_gradient)

        # Calculate bias gradients
        u_grad_bias = {user: g_loss + self.l2 * self.u_biases[user]}
        i_grad_bias = {item: g_loss + self.l2 * self.i_biases[item]}

        # Update biases
        self.u_biases = self.u_optimizer.update_after_pred(self.u_biases, u_grad_bias)
        self.i_biases = self.i_optimizer.update_after_pred(self.i_biases, i_grad_bias)

        return self



import itertools
import random
import types
import typing


def shuffle(stream: typing.Iterator, buffer_size: int, seed: int = None):
    """Shuffles a stream of data.

    This works by maintaining a buffer of elements. The first `buffer_size` elements are stored in
    memory. Once the buffer is full, a random element inside the buffer is yielded. Everytime an
    element is yielded, the next element in the stream replaces it and the buffer is sampled again.
    Increasing `buffer_size` will improve the quality of the shuffling.

    If you really want to stream over your dataset in a "good" random order, the best way is to
    split your dataset into smaller datasets and loop over them in a round-robin fashion. You may
    do this by using the ``roundrobin`` recipe from the `itertools` module.

    Parameters:
        stream: The stream to shuffle.
        buffer_size: The size of the buffer which contains the elements help in memory. Increasing
            this will increase randomness but will incur more memory usage.
        seed: Random seed used for sampling.

    Example:

        >>> from creme import stream

        >>> for i in stream.shuffle(range(15), buffer_size=5, seed=42):
        ...     print(i)
        0
        5
        2
        1
        8
        9
        6
        4
        11
        12
        10
        7
        14
        13
        3

    References:
        1. [Visualizing TensorFlow's streaming shufflers](http://www.moderndescartes.com/essays/shuffle_viz/)

    """

    rng = random.Random(seed)

    # If stream is not a generator, then we coerce it to one
    if not isinstance(stream, types.GeneratorType):
        stream = iter(stream)

    # Initialize the buffer with the first buffer_size elements of the stream
    buffer = list(itertools.islice(stream, buffer_size))

    # Deplete the stream until it is empty
    for element in stream:

        # Pick a random element from the buffer and yield it
        i = rng.randint(0, len(buffer) - 1)
        yield buffer[i]

        # Replace the yielded element from the buffer with the new element from the stream
        buffer[i] = element

    # Shuffle the remaining buffer elements and yield them one by one
    rng.shuffle(buffer)
    yield from buffer



from . import base


class SMSSpam(base.RemoteDataset):
    """SMS Spam Collection dataset.

    The data contains 5,574 items and 1 feature (i.e. SMS body). Spam messages represent
    13.4% of the dataset. The goal is to predict whether an SMS is a spam or not.

    References:
        1. [Almeida, T.A., Hidalgo, J.M.G. and Yamakami, A., 2011, September. Contributions to the study of SMS spam filtering: new collection and results. In Proceedings of the 11th ACM symposium on Document engineering (pp. 259-262).](http://www.dt.fee.unicamp.br/~tiago/smsspamcollection/doceng11.pdf)

    """

    def __init__(self):
        super().__init__(
            n_samples=5_574,
            n_features=1,
            task=base.BINARY_CLF,
            url='https://archive.ics.uci.edu/ml/machine-learning-databases/00228/smsspamcollection.zip',
            size=477907,
            filename='SMSSpamCollection'
        )

    def _iter(self):
        with open(self.path) as f:
            for row in f:
                label, body = row.split('\t')
                yield ({'body': body}, label == 'spam')



"""Model evaluation and selection."""
import bisect
import collections
import datetime as dt
import math
import time
import typing

from creme import base
from creme import metrics
from creme import utils
from creme import stream


__all__ = ['progressive_val_score']


def progressive_val_score(X_y: base.typing.Stream, model: base.Predictor, metric: metrics.Metric,
                          moment: typing.Union[str, typing.Callable] = None,
                          delay: typing.Union[str, int, dt.timedelta, typing.Callable] = None,
                          print_every=0, show_time=False, show_memory=False,
                          **print_kwargs) -> metrics.Metric:
    """Evaluates the performance of a model on a streaming dataset.

    This method is the canonical way to evaluate a model's performance. When used correctly, it
    allows you to exactly assess how a model would have performed in a production scenario.

    `X_y` is converted into a stream of questions and answers. At each step the model is either
    asked to predict an observation, or is either updated. The target is only revealed to the model
    after a certain amount of time, which is determined by the `delay` parameter. Note that under
    the hood this uses the `stream.simulate_qa` function to go through the data in arrival order.

    By default, there is no delay, which means that the samples are processed one after the other.
    When there is no delay, this function essentially performs progressive validation. When there
    is a delay, then we refer to it as delayed progressive validation.

    It is recommended to use this method when you want to determine a model's performance on a
    dataset. In particular, it is advised to use the `delay` parameter in order to get a reliable
    assessment. Indeed, in a production scenario, it is often the case that ground truths are made
    available after a certain amount of time. By using this method, you can reproduce this scenario
    and therefore truthfully assess what would have been the performance of a model on a given
    dataset.

    Parameters:
        X_y: The stream of observations against which the model will be evaluated.
        model: The model to evaluate.
        metric: The metric used to evaluate the model's predictions.
        moment: The attribute used for measuring time. If a callable is passed, then it is expected
            to take as input a `dict` of features. If `None`, then the observations are implicitely
            timestamped in the order in which they arrive.
        delay: The amount to wait before revealing the target associated with each observation to
            the model. This value is expected to be able to sum with the `moment` value. For
            instance, if `moment` is a `datetime.date`, then `delay` is expected to be a
            `datetime.timedelta`. If a callable is passed, then it is expected to take as input a
            `dict` of features and the target. If a `str` is passed, then it will be used to access
            the relevant field from the features. If `None` is passed, then no delay will be
            used, which leads to doing standard online validation.
        print_every: Iteration number at which to print the current metric. This only takes into
            account the predictions, and not the training steps.
        show_time: Whether or not to display the elapsed time.
        show_memory: Whether or not to display the memory usage of the model.
        print_kwargs: Extra keyword arguments are passed to the `print` function. For instance,
            this allows providing a `file` argument, which indicates where to output progress.

    Example:

        Take the following model:

        >>> from creme import linear_model
        >>> from creme import preprocessing

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression()
        ... )

        We can evaluate it on the `Phishing` dataset as so:

        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import model_selection

        >>> model_selection.progressive_val_score(
        ...     model=model,
        ...     X_y=datasets.Phishing(),
        ...     metric=metrics.ROCAUC(),
        ...     print_every=200
        ... )
        [200] ROCAUC: 0.897995
        [400] ROCAUC: 0.920896
        [600] ROCAUC: 0.931339
        [800] ROCAUC: 0.939909
        [1,000] ROCAUC: 0.947417
        [1,200] ROCAUC: 0.950304
        ROCAUC: 0.950363

        We haven't specified a delay, therefore this is strictly equivalent to the following piece
        of code:

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression()
        ... )

        >>> metric = metrics.ROCAUC()

        >>> for x, y in datasets.Phishing():
        ...     y_pred = model.predict_proba_one(x)
        ...     metric = metric.update(y, y_pred)
        ...     model = model.fit_one(x, y)

        >>> metric
        ROCAUC: 0.950363

        When `print_every` is specified, the current state is printed at regular intervals. Under
        the hood, Python's `print` method is being used. You can pass extra keyword arguments to
        modify its behavior. For instance, you may use the `file` argument if you want to log the
        progress to a file of your choice.

        >>> with open('progress.log', 'w') as f:
        ...     metric = model_selection.progressive_val_score(
        ...         model=model,
        ...         X_y=datasets.Phishing(),
        ...         metric=metrics.ROCAUC(),
        ...         print_every=200,
        ...         file=f
        ...     )

        >>> with open('progress.log') as f:
        ...     for line in f.read().splitlines():
        ...         print(line)
        [200] ROCAUC: 0.94
        [400] ROCAUC: 0.946969
        [600] ROCAUC: 0.9517
        [800] ROCAUC: 0.954238
        [1,000] ROCAUC: 0.958207
        [1,200] ROCAUC: 0.96002

        Note that the performance is slightly better than above because we haven't used a fresh
        copy of the model. Instead, we've reused the existing model which has already done a full
        pass on the data.

        >>> import os; os.remove('progress.log')

    References:
        1. [Beating the Hold-Out: Bounds for K-fold and Progressive Cross-Validation](http://hunch.net/~jl/projects/prediction_bounds/progressive_validation/coltfinal.pdf)
        2. [Grzenda, M., Gomes, H.M. and Bifet, A., 2019. Delayed labelling evaluation for data streams. Data Mining and Knowledge Discovery, pp.1-30](https://link.springer.com/content/pdf/10.1007%2Fs10618-019-00654-y.pdf)

    """

    # Check that the model and the metric are in accordance
    if not metric.works_with(model):
        raise ValueError(f'{metric.__class__.__name__} metric is not compatible with {model}')

    # Determine if predict_one or predict_proba_one should be used in case of a classifier
    pred_func = model.predict_one
    is_classifier = isinstance(utils.estimator_checks.guess_model(model), base.Classifier)
    if is_classifier and not metric.requires_labels:
        pred_func = model.predict_proba_one

    preds = {}

    n_total_answers = 0
    if show_time:
        start = time.perf_counter()

    for i, x, y in stream.simulate_qa(X_y, moment, delay, copy=True):

        # Question
        if y is None:
            preds[i] = pred_func(x=x)
            continue

        # Answer
        y_pred = preds.pop(i)
        if y_pred != {} and y_pred is not None:
            metric.update(y_true=y, y_pred=y_pred)
        model.fit_one(x=x, y=y)

        # Update the answer counter
        n_total_answers += 1
        if print_every and not n_total_answers % print_every:
            msg = f'[{n_total_answers:,d}] {metric}'
            if show_time:
                now = time.perf_counter()
                msg += f' – {dt.timedelta(seconds=int(now - start))}'
            if show_memory:
                msg += f' – {model._memory_usage}'
            print(msg, **print_kwargs)

    return metric



import typing

from creme import base


__all__ = ['Discard', 'Select', 'SelectType']


class Discard(base.Transformer):
    """Removes features according to a blacklist.

    This can be used in a pipeline when you want to remove certain features. The `transform_one`
    method is pure, and therefore returns a fresh new dictionary instead of removing the specified
    keys from the input.

    Parameters:
        blacklist: Key(s) to discard.

    Example:

        >>> from creme import compose

        >>> x = {'a': 42, 'b': 12, 'c': 13}
        >>> compose.Discard('a', 'b').transform_one(x)
        {'c': 13}

        You can chain a discarder with any estimator in order to apply said estimator to the
        desired features.

        >>> from creme import preprocessing

        >>> x = {'sales': 10, 'shop': 'Ikea', 'country': 'Sweden'}

        >>> pipeline = (
        ...     compose.Discard('shop', 'country') |
        ...     preprocessing.PolynomialExtender()
        ... )
        >>> pipeline.transform_one(x)
        {'sales': 10, 'sales*sales': 100}

    """

    def __init__(self, *blacklist: typing.Tuple[base.typing.FeatureName]):
        self.blacklist = set(blacklist)

    def transform_one(self, x):
        return {i: xi for i, xi in x.items() if i not in self.blacklist}

    def __str__(self):
        return '~' + str(sorted(self.blacklist))

    def __repr__(self):
        if self.blacklist:
            return 'Discard (\n  ' + '\n  '.join(map(str, sorted(self.blacklist))) + '\n)'
        return 'Discard ()'

    def _set_params(self, blacklist=None):
        if not blacklist:
            blacklist = self.blacklist
        return Select(*blacklist)


class Select(base.Transformer):
    """Selects features according to a whitelist.

    This can be used in a pipeline when you want to remove certain features. The `transform_one`
    method is pure, and therefore returns a fresh new dictionary instead of removing the specified
    keys from the input.

    Parameters:
        whitelist: Key(s) to keep.

    Example:

        >>> from creme import compose

        >>> x = {'a': 42, 'b': 12, 'c': 13}
        >>> compose.Select('c').transform_one(x)
        {'c': 13}

        You can chain a selector with any estimator in order to apply said estimator to the
        desired features.

        >>> from creme import preprocessing

        >>> x = {'sales': 10, 'shop': 'Ikea', 'country': 'Sweden'}

        >>> pipeline = (
        ...     compose.Select('sales') |
        ...     preprocessing.PolynomialExtender()
        ... )
        >>> pipeline.transform_one(x)
        {'sales': 10, 'sales*sales': 100}

    """

    def __init__(self, *whitelist: typing.Tuple[base.typing.FeatureName]):
        self.whitelist = set(whitelist)

    def transform_one(self, x):
        return {i: x[i] for i in self.whitelist}

    def __str__(self):
        return str(sorted(self.whitelist))

    def __repr__(self):
        if self.whitelist:
            return 'Select (\n  ' + '\n  '.join(map(str, sorted(self.whitelist))) + '\n)'
        return 'Select ()'

    def _get_params(self):
        return self.whitelist

    def _set_params(self, whitelist=None):
        if not whitelist:
            whitelist = self.whitelist
        return Select(*whitelist)


class SelectType(base.Transformer):
    """Selects features based on their type.

    This is practical when you want to apply different preprocessing steps to different kinds of
    features. For instance, a common usecase is to apply a `preprocessing.StandardScaler` to
    numeric features and a `preprocessing.OneHotEncoder` to categorical features.

    Parameters:
        types: Python types which you want to select. Under the hood, the `isinstance` method
            will be used to check if a value is of a given type.

    Example:

        >>> import numbers
        >>> from creme import compose
        >>> from creme import linear_model
        >>> from creme import preprocessing

        >>> num = compose.SelectType(numbers.Number) | preprocessing.StandardScaler()
        >>> cat = compose.SelectType(str) | preprocessing.OneHotEncoder()
        >>> model = (num + cat) | linear_model.LogisticRegression()

    """

    def __init__(self, *types: typing.Tuple[type]):
        self.types = types

    def transform_one(self, x):
        return {i: xi for i, xi in x.items() if isinstance(xi, self.types)}

    def __str__(self):
        return f'Select({", ".join(t.__name__ for t in self.types)})'

    def __repr__(self):
        if self.types:
            return 'Select (\n  ' + '\n  '.join(map(str, sorted(self.types))) + '\n)'
        return 'Select ()'

    def _get_params(self):
        return self.types

    def _set_params(self, types=None):
        if not types:
            types = self.types
        return SelectType(*types)



import collections

import numpy as np

from .. import base
from .. import optim
from .. import utils


__all__ = ['PAClassifier', 'PARegressor']


class BasePA:

    def __init__(self, C, mode, fit_intercept):
        self.C = C
        self.mode = mode
        self.calc_tau = {0: self._calc_tau_0, 1: self._calc_tau_1, 2: self._calc_tau_2}[mode]
        self.fit_intercept = fit_intercept
        self.weights = collections.defaultdict(float)
        self.intercept = 0.

    @classmethod
    def _calc_tau_0(cls, x, loss):
        norm = utils.math.norm(x, order=2) ** 2
        if norm > 0:
            return loss / utils.math.norm(x, order=2) ** 2
        return 0

    def _calc_tau_1(self, x, loss):
        norm = utils.math.norm(x, order=2) ** 2
        if norm > 0:
            return min(self.C, loss / norm)
        return 0

    def _calc_tau_2(self, x, loss):
        return loss / (utils.math.norm(x, order=2) ** 2 + 0.5 / self.C)


class PARegressor(BasePA, base.Regressor):
    """Passive-aggressive learning for regression.

    Parameters:
        C
        mode
        eps
        fit_intercept

    Example:

        The following example is taken from [this blog post](https://www.bonaccorso.eu/2017/10/06/ml-algorithms-addendum-passive-aggressive-algorithms/).

        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import stream
        >>> import numpy as np
        >>> from sklearn import datasets

        >>> np.random.seed(1000)
        >>> X, y = datasets.make_regression(n_samples=500, n_features=4)

        >>> model = linear_model.PARegressor(
        ...     C=0.01,
        ...     mode=2,
        ...     eps=0.1,
        ...     fit_intercept=False
        ... )
        >>> metric = metrics.MAE() + metrics.MSE()

        >>> for xi, yi in stream.iter_array(X, y):
        ...     y_pred = model.predict_one(xi)
        ...     model = model.fit_one(xi, yi)
        ...     metric = metric.update(yi, y_pred)

        >>> print(metric)
        MAE: 9.809402, MSE: 472.393532

    References:
        1. [Crammer, K., Dekel, O., Keshet, J., Shalev-Shwartz, S. and Singer, Y., 2006. Online passive-aggressive algorithms. Journal of Machine Learning Research, 7(Mar), pp.551-585.](http://jmlr.csail.mit.edu/papers/volume7/crammer06a/crammer06a.pdf)

    """

    def __init__(self, C=1.0, mode=1, eps=0.1, fit_intercept=True):
        super().__init__(C=C, mode=mode, fit_intercept=fit_intercept)
        self.eps = eps
        self.loss = optim.losses.EpsilonInsensitiveHinge(eps=eps)

    def fit_one(self, x, y):

        y_pred = self.predict_one(x)
        tau = self.calc_tau(x, self.loss(y, y_pred))
        step = tau * np.sign(y - y_pred)

        for i, xi in x.items():
            self.weights[i] += step * xi
        if self.fit_intercept:
            self.intercept += step

        return self

    def predict_one(self, x):
        return utils.math.dot(x, self.weights) + self.intercept


class PAClassifier(BasePA, base.BinaryClassifier):
    """Passive-aggressive learning for classification.

    Parameters:
        C
        mode
        eps
        fit_intercept

    Example:

        The following example is taken from [this blog post](https://www.bonaccorso.eu/2017/10/06/ml-algorithms-addendum-passive-aggressive-algorithms/).

        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import stream
        >>> import numpy as np
        >>> from sklearn import datasets
        >>> from sklearn import model_selection

        >>> np.random.seed(1000)
        >>> X, y = datasets.make_classification(
        ...     n_samples=5000,
        ...     n_features=4,
        ...     n_informative=2,
        ...     n_redundant=0,
        ...     n_repeated=0,
        ...     n_classes=2,
        ...     n_clusters_per_class=2
        ... )

        >>> X_train, X_test, y_train, y_test = model_selection.train_test_split(
        ...     X,
        ...     y,
        ...     test_size=0.35,
        ...     random_state=1000
        ... )

        >>> model = linear_model.PAClassifier(
        ...     C=0.01,
        ...     mode=1
        ... )

        >>> for xi, yi in stream.iter_array(X_train, y_train):
        ...     y_pred = model.fit_one(xi, yi)

        >>> metric = metrics.Accuracy() + metrics.LogLoss()

        >>> for xi, yi in stream.iter_array(X_test, y_test):
        ...     metric = metric.update(yi, model.predict_proba_one(xi))

        >>> print(metric)
        Accuracy: 88.46%, LogLoss: 0.325727

    References:
        1. [Crammer, K., Dekel, O., Keshet, J., Shalev-Shwartz, S. and Singer, Y., 2006. Online passive-aggressive algorithms. Journal of Machine Learning Research, 7(Mar), pp.551-585](http://jmlr.csail.mit.edu/papers/volume7/crammer06a/crammer06a.pdf)

    """

    def __init__(self, C=1., mode=1, fit_intercept=True):
        super().__init__(C=C, mode=mode, fit_intercept=fit_intercept)
        self.loss = optim.losses.Hinge()

    def fit_one(self, x, y):

        y_pred = utils.math.dot(x, self.weights) + self.intercept
        tau = self.calc_tau(x, self.loss(y, y_pred))
        step = tau * (y or -1)  # y == False becomes -1

        for i, xi in x.items():
            self.weights[i] += step * xi
        if self.fit_intercept:
            self.intercept += step

        return self

    def predict_proba_one(self, x):
        y_pred = utils.math.sigmoid(utils.math.dot(x, self.weights) + self.intercept)
        return {False: 1. - y_pred, True: y_pred}



import collections
import copy
import functools
import typing

from creme import base
from creme import stats


__all__ = ['Agg', 'TargetAgg']


class Agg(base.Transformer):
    """Computes a streaming aggregate.

    This transformer allows to compute an aggregate statistic, very much like the groupby method
    from `pandas`, but on a streaming dataset. This makes use of the streaming statistics from the
    `stats` module.

    When `fit_one` is called, the running statistic `how` of group `by` is updated with the value
    of `on`. Meanwhile, the output of `transform_one` is a single-element dictionary, where the key
    is the name of the aggregate and the value is the current value of the statistic for the
    relevant group. The key is automatically inferred from the parameters.

    Note that you can use a `compose.TransformerUnion` to extract many aggregate statistics in a
    concise manner.

    Parameters:
        on: The feature on which to compute the aggregate statistic.
        by: The feature by which to group the data.
        how: The statistic to compute.

    Attributes:
        groups (collections.defaultdict): Maps group keys to univariate statistics.
        feature_name (str): The name of the feature used in the output.

    Example:

        Consider the following dataset:

        >>> X = [
        ...     {'country': 'France', 'place': 'Taco Bell', 'revenue': 42},
        ...     {'country': 'Sweden', 'place': 'Burger King', 'revenue': 16},
        ...     {'country': 'France', 'place': 'Burger King', 'revenue': 24},
        ...     {'country': 'Sweden', 'place': 'Taco Bell', 'revenue': 58},
        ...     {'country': 'Sweden', 'place': 'Burger King', 'revenue': 20},
        ...     {'country': 'France', 'place': 'Taco Bell', 'revenue': 50},
        ...     {'country': 'France', 'place': 'Burger King', 'revenue': 10},
        ...     {'country': 'Sweden', 'place': 'Taco Bell', 'revenue': 80}
        ... ]

        As an example, we can calculate the average (how) revenue (on) for each place (by):

        >>> from creme import feature_extraction
        >>> from creme import stats

        >>> agg = feature_extraction.Agg(
        ...     on='revenue',
        ...     by='place',
        ...     how=stats.Mean()
        ... )

        >>> for x in X:
        ...     agg = agg.fit_one(x)
        ...     print(agg.transform_one(x))
        {'revenue_mean_by_place': 42.0}
        {'revenue_mean_by_place': 16.0}
        {'revenue_mean_by_place': 20.0}
        {'revenue_mean_by_place': 50.0}
        {'revenue_mean_by_place': 20.0}
        {'revenue_mean_by_place': 50.0}
        {'revenue_mean_by_place': 17.5}
        {'revenue_mean_by_place': 57.5}

        You can compute an aggregate over multiple keys by passing a tuple to the `by` argument.
        For instance, we can compute the maximum (how) revenue (on) per place as well as per
        day (by):

        >>> agg = feature_extraction.Agg(
        ...     on='revenue',
        ...     by=['place', 'country'],
        ...     how=stats.Max()
        ... )

        >>> for x in X:
        ...     agg = agg.fit_one(x)
        ...     print(agg.transform_one(x))
        {'revenue_max_by_place_and_country': 42}
        {'revenue_max_by_place_and_country': 16}
        {'revenue_max_by_place_and_country': 24}
        {'revenue_max_by_place_and_country': 58}
        {'revenue_max_by_place_and_country': 20}
        {'revenue_max_by_place_and_country': 50}
        {'revenue_max_by_place_and_country': 24}
        {'revenue_max_by_place_and_country': 80}

        You can use a `compose.TransformerUnion` in order to calculate multiple aggregates in one
        go. The latter can be constructed by using the `+` operator:

        >>> agg = (
        ...     feature_extraction.Agg(on='revenue', by='place', how=stats.Mean()) +
        ...     feature_extraction.Agg(on='revenue', by=['place', 'country'], how=stats.Max())
        ... )

        >>> import pprint
        >>> for x in X:
        ...     agg = agg.fit_one(x)
        ...     pprint.pprint(agg.transform_one(x))
        {'revenue_max_by_place_and_country': 42, 'revenue_mean_by_place': 42.0}
        {'revenue_max_by_place_and_country': 16, 'revenue_mean_by_place': 16.0}
        {'revenue_max_by_place_and_country': 24, 'revenue_mean_by_place': 20.0}
        {'revenue_max_by_place_and_country': 58, 'revenue_mean_by_place': 50.0}
        {'revenue_max_by_place_and_country': 20, 'revenue_mean_by_place': 20.0}
        {'revenue_max_by_place_and_country': 50, 'revenue_mean_by_place': 50.0}
        {'revenue_max_by_place_and_country': 24, 'revenue_mean_by_place': 17.5}
        {'revenue_max_by_place_and_country': 80, 'revenue_mean_by_place': 57.5}

    References:
        1. [Streaming groupbys in pandas for big datasets](https://maxhalford.github.io/blog/streaming-groupbys-in-pandas-for-big-datasets/)

    """

    def __init__(self, on: str, by: typing.Union[str, typing.List[str]], how: stats.Univariate):
        self.on = on
        self.by = by if isinstance(by, list) else [by]
        self.how = how
        self.groups = collections.defaultdict(functools.partial(copy.deepcopy, how))
        self.feature_name = f'{self.on}_{self.how.name}_by_{"_and_".join(self.by)}'

    def _get_key(self, x):
        return '_'.join(str(x[k]) for k in self.by)

    def fit_one(self, x):
        self.groups[self._get_key(x)].update(x[self.on])
        return self

    def transform_one(self, x):
        return {self.feature_name: self.groups[self._get_key(x)].get()}

    def __str__(self):
        return self.feature_name


class TargetAgg(base.SupervisedTransformer):
    """Computes a streaming aggregate of the target values.

    This transformer is identical to `feature_extraction.Agg`, the only difference is that it
    operates on the target rather than on a feature. At each step, the running statistic `how` of
    target values in group `by` is updated with the target. It is therefore a supervised
    transformer.

    Parameters:
        by: The feature by which to group the target values.
        how: The statistic to compute.
        target_name: The target name which is used in the result.

    Attributes:
        groups (dict): Maps group keys to univariate statistics.
        feature_name (str): The name of the feature in the output.

    Example:

        Consider the following dataset, where the second value of each value is the target:

        >>> X_y = [
        ...     ({'country': 'France', 'place': 'Taco Bell'}, 42),
        ...     ({'country': 'Sweden', 'place': 'Burger King'}, 16),
        ...     ({'country': 'France', 'place': 'Burger King'}, 24),
        ...     ({'country': 'Sweden', 'place': 'Taco Bell'}, 58),
        ...     ({'country': 'Sweden', 'place': 'Burger King'}, 20),
        ...     ({'country': 'France', 'place': 'Taco Bell'}, 50),
        ...     ({'country': 'France', 'place': 'Burger King'}, 10),
        ...     ({'country': 'Sweden', 'place': 'Taco Bell'}, 80)
        ... ]

        As an example, let's perform a target encoding of the `place` feature. Instead of simply
        updating a running average, we use a `stats.BayesianMean` which allows us to incorporate
        some prior knowledge. This makes subsequent models less prone to overfitting. Indeed, it
        dampens the fact that too few samples might have been seen within a group.

        >>> from creme import feature_extraction
        >>> from creme import stats

        >>> agg = feature_extraction.TargetAgg(
        ...     by='place',
        ...     how=stats.BayesianMean(
        ...         prior=3,
        ...         prior_weight=1
        ...     )
        ... )

        >>> for x, y in X_y:
        ...     print(agg.transform_one(x))
        ...     agg = agg.fit_one(x, y)
        {'target_bayes_mean_by_place': 3.0}
        {'target_bayes_mean_by_place': 3.0}
        {'target_bayes_mean_by_place': 9.5}
        {'target_bayes_mean_by_place': 22.5}
        {'target_bayes_mean_by_place': 14.333}
        {'target_bayes_mean_by_place': 34.333}
        {'target_bayes_mean_by_place': 15.75}
        {'target_bayes_mean_by_place': 38.25}

        Just like with `feature_extraction.Agg`, we can specify multiple features on which to
        group the data:

        >>> agg = feature_extraction.TargetAgg(
        ...     by=['place', 'country'],
        ...     how=stats.BayesianMean(
        ...         prior=3,
        ...         prior_weight=1
        ...     )
        ... )

        >>> for x, y in X_y:
        ...     print(agg.transform_one(x))
        ...     agg = agg.fit_one(x, y)
        {'target_bayes_mean_by_place_and_country': 3.0}
        {'target_bayes_mean_by_place_and_country': 3.0}
        {'target_bayes_mean_by_place_and_country': 3.0}
        {'target_bayes_mean_by_place_and_country': 3.0}
        {'target_bayes_mean_by_place_and_country': 9.5}
        {'target_bayes_mean_by_place_and_country': 22.5}
        {'target_bayes_mean_by_place_and_country': 13.5}
        {'target_bayes_mean_by_place_and_country': 30.5}

    References:
        1. [Streaming groupbys in pandas for big datasets](https://maxhalford.github.io/blog/streaming-groupbys-in-pandas-for-big-datasets/)

    """

    def __init__(self, by: typing.Union[str, typing.List[str]], how: stats.Univariate,
                 target_name='target'):
        self.by = by if isinstance(by, list) else [by]
        self.how = how
        self.target_name = target_name
        self.groups = collections.defaultdict(functools.partial(copy.deepcopy, how))
        self.feature_name = f'{target_name}_{how.name}_by_{"_and_".join(self.by)}'

    def _get_key(self, x):
        return '_'.join(str(x[k]) for k in self.by)

    def fit_one(self, x, y):
        self.groups[self._get_key(x)].update(y)
        return self

    def transform_one(self, x):
        return {self.feature_name: self.groups[self._get_key(x)].get()}

    def __str__(self):
        return self.feature_name



import collections


class Skyline(collections.UserList):
    """A skyline is set of points which is not dominated by any other point.

    This implementation uses a block nested loop. Identical observations are all part of the
    skyline if applicable.

    Parameters:
        minimize: A list of features for which the values need to be minimized. Can be omitted as
            long as `maximize` is specified.
        maximize: A list of features for which the values need to be maximized. Can be omitted as
            long as `minimize` is specified.

    Example:

        Here is an example taken from [this](https://maxhalford.github.io/blog/skyline-queries-in-python) blog post.

        >>> import random
        >>> from creme import utils
        >>> import matplotlib.pyplot as plt

        >>> city_prices = {
        ...     'Bordeaux': 4045,
        ...     'Lyon': 4547,
        ...     'Toulouse': 3278
        ... }

        >>> def random_house():
        ...     city = random.choice(['Bordeaux', 'Lyon', 'Toulouse'])
        ...     size = round(random.gauss(200, 50))
        ...     price = round(random.uniform(0.8, 1.2) * city_prices[city] * size)
        ...     return {'city': city, 'size': size, 'price': price}

        >>> skyline = utils.Skyline(minimize=['price'], maximize=['size'])

        >>> random.seed(42)

        >>> for _ in range(100):
        ...     house = random_house()
        ...     skyline = skyline.update(house)

        >>> print(len(skyline))
        13

        >>> print(skyline[0])
        {'city': 'Toulouse', 'size': 280, 'price': 763202}

        >>> fig, ax = plt.subplots()
        >>> scatter = ax.scatter(
        ...     x=[h['size'] for h in skyline],
        ...     y=[h['price'] for h in skyline]
        ... )
        >>> grid = ax.grid()
        >>> title = ax.set_title('Houses skyline')
        >>> xlabel = ax.set_xlabel('Size')
        >>> ylabel = ax.set_ylabel('Price')

        .. image:: ../../docs/img/skyline_docstring.svg
            :align: center

        Here is another example using the kart data from *Mario Kart: Double Dash!!*.

        >>> import collections
        >>> from creme import utils

        >>> Kart = collections.namedtuple(
        ...      'Kart',
        ...      'name speed off_road acceleration weight turbo'
        ... )

        >>> karts = [
        ...     Kart('Red Fire', 5, 4, 4, 5, 2),
        ...     Kart('Green Fire', 7, 3, 3, 4, 2),
        ...     Kart('Heart Coach', 4, 6, 6, 5, 2),
        ...     Kart('Bloom Coach', 6, 4, 5, 3, 2),
        ...     Kart('Turbo Yoshi', 4, 5, 6, 6, 2),
        ...     Kart('Turbo Birdo', 6, 4, 4, 7, 2),
        ...     Kart('Goo-Goo Buggy', 1, 9, 9, 2, 3),
        ...     Kart('Rattle Buggy', 2, 9, 8, 2, 3),
        ...     Kart('Toad Kart', 3, 9, 7, 2, 3),
        ...     Kart('Toadette Kart', 1, 9, 9, 2, 3),
        ...     Kart('Koopa Dasher', 2, 8, 8, 3, 3),
        ...     Kart('Para-Wing', 1, 8, 9, 3, 3),
        ...     Kart('DK Jumbo', 8, 2, 2, 8, 1),
        ...     Kart('Barrel Train', 8, 7, 3, 5, 3),
        ...     Kart('Koopa King', 9, 1, 1, 9, 1),
        ...     Kart('Bullet Blaster', 8, 1, 4, 1, 3),
        ...     Kart('Wario Car', 7, 3, 3, 7, 1),
        ...     Kart('Waluigi Racer', 5, 9, 5, 6, 2),
        ...     Kart('Piranha Pipes', 8, 7, 2, 9, 1),
        ...     Kart('Boo Pipes', 2, 9, 8, 9, 1),
        ...     Kart('Parade Kart', 7, 3, 4, 7, 3)
        ... ]

        >>> skyline = utils.Skyline(
        ...     maximize=['speed', 'off_road', 'acceleration', 'turbo'],
        ...     minimize=['weight']
        ... )

        >>> for kart in karts:
        ...     skyline = skyline.update(kart._asdict())

        >>> best_cart_names = [kart['name'] for kart in skyline]
        >>> for name in best_cart_names:
        ...     print(f'- {name}')
        - Green Fire
        - Heart Coach
        - Bloom Coach
        - Goo-Goo Buggy
        - Rattle Buggy
        - Toad Kart
        - Toadette Kart
        - Barrel Train
        - Koopa King
        - Bullet Blaster
        - Waluigi Racer
        - Parade Kart

        >>> for name in sorted(set(kart.name for kart in karts) - set(best_cart_names)):
        ...     print(f'- {name}')
        - Boo Pipes
        - DK Jumbo
        - Koopa Dasher
        - Para-Wing
        - Piranha Pipes
        - Red Fire
        - Turbo Birdo
        - Turbo Yoshi
        - Wario Car

    References:
        1. [Skyline queries in Python](https://maxhalford.github.io/blog/skyline-queries-in-python)
        2. [Borzsony, S., Kossmann, D. and Stocker, K., 2001, April. The skyline operator. In Proceedings 17th international conference on data engineering (pp. 421-430). IEEE.](https://infolab.usc.edu/csci599/Fall2007/papers/e-1.pdf)
        3. [Tao, Y. and Papadias, D., 2006. Maintaining sliding window skylines on data streams. IEEE Transactions on Knowledge and Data Engineering, 18(3), pp.377-391.](http://www.cs.ust.hk/~dimitris/PAPERS/TKDE06-Sky.pdf)

    """

    def __init__(self, minimize: list = None, maximize: list = None):

        super().__init__()

        self.minimize = [] if minimize is None else minimize
        self.maximize = [] if maximize is None else maximize

        if len(self.minimize) + len(self.maximize) == 0:
            raise ValueError('At least one name has to be specified')

    def _count_diffs(self, a, b):
        n_better = 0
        n_worse = 0

        for f in self.minimize:
            n_better += a[f] < b[f]
            n_worse += a[f] > b[f]

        for f in self.maximize:
            n_better += a[f] > b[f]
            n_worse += a[f] < b[f]

        return n_better, n_worse

    def update(self, x):

        # If the skyline is empty then the first element is part of the skyline
        if not self:
            self.append(x)
            return self

        to_drop = []
        is_dominated = False

        for i, _ in enumerate(self):

            n_better, n_worse = self._count_diffs(x, self[i])

            # x is dominated by one element in the skyline
            if n_worse > 0 and n_better == 0:
                is_dominated = True
                break

            # x dominates one element in the skyline
            if n_better > 0 and n_worse == 0:
                to_drop.append(i)

        if is_dominated:
            return self

        # Remove dominated elements
        if to_drop:
            for i in sorted(to_drop, reverse=True):
                del self[i]

        # Add x to the skyline
        self.append(x)

        return self



"""Naive Bayes algorithms."""
from .bernoulli import BernoulliNB
from .complement import ComplementNB
from .gaussian import GaussianNB
from .multinomial import MultinomialNB


__all__ = [
    'BernoulliNB',
    'ComplementNB',
    'GaussianNB',
    'MultinomialNB'
]



"""Utility classes and functions."""
from . import estimator_checks
from . import math
from . import pretty
from .histogram import Histogram
from .sdft import SDFT
from .skyline import Skyline
from .window import Window
from .window import SortedWindow


__all__ = [
    'estimator_checks',
    'Histogram',
    'SDFT',
    'Skyline',
    'SortedWindow',
    'Window'
]



"""Model composition.

This module contains utilities for merging multiple modeling steps into a single pipeline. Although
pipelines are not the only way to process a stream of data, we highly encourage you to use them.

"""
from .func import FuncTransformer
from .grouper import Grouper
from .pipeline import Pipeline
from .rename import Renamer
from .union import TransformerUnion
from .select import Discard
from .select import Select
from .select import SelectType


__all__ = [
    'Discard',
    'FuncTransformer',
    'Grouper',
    'Pipeline',
    'Renamer',
    'TransformerUnion',
    'Select',
    'SelectType'
]



from creme import stream

from . import base


class MovieLens100K(base.RemoteDataset):
    """MovieLens 100K dataset.

    MovieLens datasets were collected by the GroupLens Research Project at the University of
    Minnesota. This dataset consists of 100,000 ratings (1-5) from 943 users on 1682 movies. Each
    user has rated at least 20 movies. User and movie informations are provided. The data was
    collected through the MovieLens web site (movielens.umn.edu) during the seven-month period from
    September 19th, 1997 through April 22nd, 1998.

    References:
        1. [The MovieLens Datasets: History and Context](http://dx.doi.org/10.1145/2827872)

    """

    def __init__(self):
        super().__init__(
            n_samples=100_000,
            n_features=10,
            task=base.REG,
            url='https://maxhalford.github.io/files/datasets/ml_100k.zip',
            size=11057876,
            filename='ml_100k.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='rating',
            converters={
                'timestamp': int,
                'release_date': int,
                'age': float,
                'rating': float
            },
            delimiter='\t'
        )



import math

from .. import stats

from . import base


__all__ = ['Gaussian']


class Gaussian(base.ContinuousDistribution):
    """Normal distribution with parameters mu and sigma.

    Example:

        >>> from creme import proba

        >>> p = proba.Gaussian().update(6).update(7)

        >>> p
        𝒩(μ=6.500, σ=0.707)

        >>> p.pdf(6.5)
        0.564189

    """

    def __init__(self):
        self._var = stats.Var()

    @property
    def n_samples(self):
        return self._var.mean.n

    @property
    def mu(self):
        return self._var.mean.get()

    @property
    def sigma(self):
        return self._var.get() ** 0.5

    @property
    def mode(self):
        return self.mu

    def __str__(self):
        return f'𝒩(μ={self.mu:.3f}, σ={self.sigma:.3f})'

    def update(self, x):
        self._var.update(x)
        return self

    def pdf(self, x):
        var = self._var.get()
        if var:
            return math.exp((x - self.mu) ** 2 / (-2 * var)) / math.sqrt(math.tau * var)
        return 0.

    def cdf(self, x):
        try:
            return 0.5 * (1. + math.erf((x - self.mu) / (self.sigma * math.sqrt(2.))))
        except ZeroDivisionError:
            return 0.



import sqlalchemy
import typing

from creme import base


__all__ = ['iter_sql']


def iter_sql(query: typing.Union[str, sqlalchemy.sql.expression.Selectable],
             conn: sqlalchemy.engine.Connectable, target_name: str = None) -> base.typing.Stream:
    """Iterates over the results from an SQL query.

    Parameters:
        query: SQL query to be executed.
        conn: An SQLAlchemy construct which has an `execute` method. In other words you can pass an
            engine, a connection, or a session.
        target_name: The name of the target field. If this is `None`, then `y` will also be `None`.

    Example:

        As an example we'll create an in-memory database with SQLAlchemy.

        >>> import datetime as dt
        >>> import sqlalchemy

        >>> engine = sqlalchemy.create_engine('sqlite://')

        >>> metadata = sqlalchemy.MetaData()

        >>> t_sales = sqlalchemy.Table('sales', metadata,
        ...     sqlalchemy.Column('shop', sqlalchemy.String, primary_key=True),
        ...     sqlalchemy.Column('date', sqlalchemy.Date, primary_key=True),
        ...     sqlalchemy.Column('amount', sqlalchemy.Integer)
        ... )

        >>> metadata.create_all(engine)

        >>> sales = [
        ...     {'shop': 'Hema', 'date': dt.date(2016, 8, 2), 'amount': 20},
        ...     {'shop': 'Ikea', 'date': dt.date(2016, 8, 2), 'amount': 18},
        ...     {'shop': 'Hema', 'date': dt.date(2016, 8, 3), 'amount': 22},
        ...     {'shop': 'Ikea', 'date': dt.date(2016, 8, 3), 'amount': 14},
        ...     {'shop': 'Hema', 'date': dt.date(2016, 8, 4), 'amount': 12},
        ...     {'shop': 'Ikea', 'date': dt.date(2016, 8, 4), 'amount': 16}
        ... ]

        >>> with engine.connect() as conn:
        ...     _ = conn.execute(t_sales.insert(), sales)

        We can now query the database. We will set `amount` to be the target field.

        >>> from creme import stream

        >>> with engine.connect() as conn:
        ...     query = 'SELECT * FROM sales;'
        ...     X_y = stream.iter_sql(query, conn, target_name='amount')
        ...     for x, y in X_y:
        ...         print(x, y)
        {'shop': 'Hema', 'date': '2016-08-02'} 20
        {'shop': 'Ikea', 'date': '2016-08-02'} 18
        {'shop': 'Hema', 'date': '2016-08-03'} 22
        {'shop': 'Ikea', 'date': '2016-08-03'} 14
        {'shop': 'Hema', 'date': '2016-08-04'} 12
        {'shop': 'Ikea', 'date': '2016-08-04'} 16

    .. note::
        By default, SQLAlchemy prefetches results. Therefore, even though you can iterate over the
        resulting rows one by one, the results are in fact loaded in batch. You can modify this
        behavior by configuring the connection you pass to `iter_sql`. For instance, you can set
        the `stream_results` parameter to `True`, as [explained in SQLAlchemy's documentation](https://docs.sqlalchemy.org/en/13/core/connections.html#sqlalchemy.engine.Connection.execution_options). Note, however, that this
        isn't available for all database engines.

    """

    result_proxy = conn.execute(query)

    if target_name is None:
        for row in result_proxy:
            yield dict(row.items()), None
        return

    for row in result_proxy:
        x = dict(row.items())
        y = x.pop(target_name)
        yield x, y



cimport creme.stats.base
cimport creme.stats.mean

from . import base
from . import mean


cdef class Var(creme.stats.base.Univariate):
    """Running variance using Welford's algorithm.

    Parameters:
        ddof: Delta Degrees of Freedom. The divisor used in calculations is `n - ddof`, where `n`
            represents the number of seen elements.

    Attributes:
        mean (stats.Mean): The running mean.
        sos (float): The running sum of squares.

    Example:

        >>> import creme.stats

        >>> X = [3, 5, 4, 7, 10, 12]

        >>> var = creme.stats.Var()
        >>> for x in X:
        ...     print(var.update(x).get())
        0.0
        2.0
        1.0
        2.916666
        7.7
        12.56666

    References:
        1. [Wikipedia article on algorithms for calculating variance](https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Covariance)

    """

    cdef readonly long ddof
    cdef readonly creme.stats.mean.Mean mean
    cdef readonly double sigma

    def __init__(self, ddof=1):
        self.ddof = ddof
        self.mean = mean.Mean()

    cpdef Var update(self, double x, double w=1.):
        mean = self.mean.get()
        self.mean.update(x, w)
        if self.mean.n > self.ddof:
            self.sigma += w * ((x - mean) * (x - self.mean.get()) - self.sigma) / (self.mean.n - self.ddof)
        return self

    cpdef double get(self):
        return self.sigma


class RollingVar(base.RollingUnivariate):
    """Running variance over a window.

    Parameters:
        window_size (int): Size of the rolling window.
        ddof (int): Delta Degrees of Freedom for the variance.

    Attributes:
        sos (float): Sum of squares over the current window.
        rmean (stats.RollingMean)

    Example:

        >>> import creme

        >>> X = [1, 4, 2, -4, -8, 0]

        >>> rvar = creme.stats.RollingVar(ddof=1, window_size=2)
        >>> for x in X:
        ...     print(rvar.update(x).get())
        0.0
        4.5
        2.0
        18.0
        8.0
        32.0

        >>> rvar = creme.stats.RollingVar(ddof=1, window_size=3)
        >>> for x in X:
        ...     print(rvar.update(x).get())
        0.0
        4.5
        2.333333
        17.333333
        25.333333
        16.0

    """

    def __init__(self, window_size, ddof=1):
        self.ddof = ddof
        self.sos = 0
        self.rolling_mean = mean.RollingMean(window_size=window_size)

    @property
    def window_size(self):
        return self.rolling_mean.window_size

    def update(self, x):
        if len(self.rolling_mean) >= self.rolling_mean.size:
            self.sos -= self.rolling_mean[0] ** 2

        self.sos += x * x
        self.rolling_mean.update(x)
        return self

    @property
    def correction_factor(self):
        n = len(self.rolling_mean)
        if n > self.ddof:
            return n / (n - self.ddof)
        return 1

    def get(self):
        try:
            var = (self.sos / len(self.rolling_mean)) - self.rolling_mean.get() ** 2
            return self.correction_factor * var
        except ZeroDivisionError:
            return 0.



import abc

from . import estimator


class Clusterer(estimator.Estimator):
    """A clustering model."""

    @property
    def _is_supervised(self):
        return False

    @abc.abstractmethod
    def fit_one(self, x: dict) -> 'Clusterer':
        """Update the model with a set of features `x`.

        Parameters:
            x: A dictionary of features.

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> int:
        """Predicts the cluster number for a set of features `x`.

        Parameters:
            x: A dictionary of features.

        Returns:
            A cluster number.

        """



collect_ignore = []

try:
    import sklearn
except ImportError:
    collect_ignore.append('compat/sklearn.py')
    collect_ignore.append('compat/test_sklearn.py')

try:
    import sqlalchemy
except ImportError:
    collect_ignore.append('stream/iter_sql.py')
    collect_ignore.append('stream/test_sql.py')

try:
    import surprise
except ImportError:
    collect_ignore.append('reco/surprise.py')

try:
    import torch
except ImportError:
    collect_ignore.append('compat/pytorch.py')

try:
    import vaex
except ImportError:
    collect_ignore.append('stream/iter_vaex.py')



import collections
import typing

from creme import base

from .predictor import Predictor


class Ensemble(Predictor):
    """An ensemble model.

    Parameters:
        models

    """

    def __init__(self, models: typing.List[Predictor]):
        self.models = list(models)

    def __len__(self):
        return len(self.models)

    def __iter__(self):
        return iter(self.models)

    def __getitem__(self, index):
        return self.models[index]



import typing

from creme import base


__all__ = ['StackingBinaryClassifier']


class StackingBinaryClassifier(base.Ensemble, base.BinaryClassifier):
    """Stacking for binary classification.

    Parameters:
        classifiers
        meta_classifier
        include_features: Indicates whether or not the original features should be provided
            to the meta-model along with the predictions from each model.

    Example:

        >>> from creme import compose
        >>> from creme import datasets
        >>> from creme import ensemble
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()

        >>> model = compose.Pipeline(
        ...     ('scale', preprocessing.StandardScaler()),
        ...     ('stack', ensemble.StackingBinaryClassifier(
        ...         classifiers=[
        ...             linear_model.LogisticRegression(),
        ...             linear_model.PAClassifier(mode=1, C=0.01),
        ...             linear_model.PAClassifier(mode=2, C=0.01)
        ...         ],
        ...         meta_classifier=linear_model.LogisticRegression()
        ...     ))
        ... )

        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.878005

    References:
        1. [A Kaggler's Guide to Model Stacking in Practice](http://blog.kaggle.com/2016/12/27/a-kagglers-guide-to-model-stacking-in-practice/)

    """

    def __init__(self, classifiers: typing.List[base.BinaryClassifier],
                 meta_classifier: base.Classifier, include_features=True):
        super().__init__(classifiers)
        self.meta_classifier = meta_classifier
        self.include_features = include_features

    def fit_one(self, x, y):

        # Ask each model to make a prediction and then update it
        oof = {}
        for i, classifier in enumerate(self):
            oof[f'oof_{i}'] = classifier.predict_proba_one(x).get(True, 0.5)
            classifier.fit_one(x, y)

        # Optionally, add the base features
        if self.include_features:
            oof.update(x)

        # Update the meta-classifier using the predictions from the base classifiers
        self.meta_classifier.fit_one(oof, y)

        return self

    def predict_proba_one(self, x):

        oof = {
            f'oof_{i}': classifier.predict_proba_one(x).get(True, 0.5)
            for i, classifier in enumerate(self)
        }

        if self.include_features:
            oof.update(x)

        return self.meta_classifier.predict_proba_one(oof)



import itertools
import functools

from creme import base
from creme import utils


__all__ = ['PolynomialExtender']


def powerset(iterable, min_size, max_size, with_replacement=False):
    """powerset([A, B, C], 1, 2) --> (A,) (B,) (C,) (A, B) (A, C) (B, C)"""
    combiner = itertools.combinations_with_replacement \
        if with_replacement \
        else itertools.combinations
    sizes = range(min_size, max_size + 1)
    return itertools.chain.from_iterable(combiner(list(iterable), size) for size in sizes)


class PolynomialExtender(base.Transformer):
    """Polynomial feature extender.

    Generate features consisting of all polynomial combinations of the features with degree less
    than or equal to the specified degree.
    Be aware that the number of outputed features scales polynomially in the number of input
    features and exponentially in the degree. High degrees can cause overfitting.

    Parameters:
        degree: The maximum degree of the polynomial features.
        interaction_only: If `True` then only combinations that include an element at most
            once will be computed.
        include_bias: Whether or not to include a dummy feature which is always equal to 1.
        bias_name: Name to give to the bias feature.

    Example:

        >>> from creme import preprocessing

        >>> X = [
        ...     {'x': 0, 'y': 1},
        ...     {'x': 2, 'y': 3},
        ...     {'x': 4, 'y': 5}
        ... ]

        >>> poly = preprocessing.PolynomialExtender(degree=2, include_bias=True)
        >>> for x in X:
        ...     print(poly.fit_one(x).transform_one(x))
        {'x': 0, 'y': 1, 'x*x': 0, 'x*y': 0, 'y*y': 1, 'bias': 1}
        {'x': 2, 'y': 3, 'x*x': 4, 'x*y': 6, 'y*y': 9, 'bias': 1}
        {'x': 4, 'y': 5, 'x*x': 16, 'x*y': 20, 'y*y': 25, 'bias': 1}

        >>> X = [
        ...     {'x': 0, 'y': 1, 'z': 2},
        ...     {'x': 2, 'y': 3, 'z': 2},
        ...     {'x': 4, 'y': 5, 'z': 2}
        ... ]

        >>> poly = preprocessing.PolynomialExtender(degree=3, interaction_only=True)
        >>> for x in X:
        ...     print(poly.fit_one(x).transform_one(x))
        {'x': 0, 'y': 1, 'z': 2, 'x*y': 0, 'x*z': 0, 'y*z': 2, 'x*y*z': 0}
        {'x': 2, 'y': 3, 'z': 2, 'x*y': 6, 'x*z': 4, 'y*z': 6, 'x*y*z': 12}
        {'x': 4, 'y': 5, 'z': 2, 'x*y': 20, 'x*z': 8, 'y*z': 10, 'x*y*z': 40}

    """

    def __init__(self, degree=2, interaction_only=False, include_bias=False, bias_name='bias'):
        self.degree = degree
        self.interaction_only = interaction_only
        self.include_bias = include_bias
        self.bias_name = bias_name
        self.enumerate = functools.partial(
            powerset,
            min_size=1,
            max_size=degree,
            with_replacement=not interaction_only
        )

    def transform_one(self, x):
        features = {
            '*'.join(map(str, combo)): utils.math.prod(x[c] for c in combo)
            for combo in self.enumerate(x.keys())
        }
        if self.include_bias:
            features[self.bias_name] = 1
        return features



import collections
import copy
import functools

from creme import base
from creme import stats


class SelectKBest(base.SupervisedTransformer):
    """Removes all but the $k$ highest scoring features.

    Parameters:
        similarity
        k: The number of features to keep.

    Attributes:
        similarities (dict): The similarity instances used for each feature.
        leaderboard (dict): The actual similarity measures.

    Example:

        >>> from pprint import pprint
        >>> from creme import feature_selection
        >>> from creme import stats
        >>> from creme import stream
        >>> from sklearn import datasets

        >>> X, y = datasets.make_regression(
        ...     n_samples=100,
        ...     n_features=10,
        ...     n_informative=2,
        ...     random_state=42
        ... )

        >>> selector = feature_selection.SelectKBest(
        ...     similarity=stats.PearsonCorr(),
        ...     k=2
        ... )

        >>> for xi, yi, in stream.iter_array(X, y):
        ...     selector = selector.fit_one(xi, yi)

        >>> pprint(selector.leaderboard)
        Counter({9: 0.7898,
                7: 0.5444,
                8: 0.1062,
                2: 0.0638,
                4: 0.0538,
                5: 0.0271,
                1: -0.0312,
                6: -0.0657,
                3: -0.1501,
                0: -0.1895})

        >>> selector.transform_one(xi)
        {7: -1.2795, 9: -1.8408}

    """

    def __init__(self, similarity: stats.Bivariate, k=10):
        self.k = k
        self.similarity = similarity
        self.similarities = collections.defaultdict(functools.partial(copy.deepcopy, similarity))
        self.leaderboard = collections.Counter()

    def fit_one(self, x, y):

        for i, xi in x.items():
            self.leaderboard[i] = self.similarities[i].update(xi, y).get()

        return self

    def transform_one(self, x):

        best_features = set(pair[0] for pair in self.leaderboard.most_common(self.k))

        if self.leaderboard:

            return {
                i: xi
                for i, xi in x.items()
                if i in best_features
            }

        return copy.deepcopy(x)



import collections
import math

from . import base


__all__ = ['MultinomialNB']


class MultinomialNB(base.BaseNB):
    """Naive Bayes classifier for multinomial models.

    The input vector has to contain positive values, such as counts or TF-IDF values.

    This class inherits `predict_proba_one` from `naive_bayes.BaseNB` which itself inherits
    `predict_one` from `base.MultiClassifier`.

    Parameters:
        alpha: Additive (Laplace/Lidstone) smoothing parameter (use 0 for no smoothing).

    Attributes:
        class_dist (proba.Multinomial): Class prior probability distribution.
        feature_counts (collections.defaultdict): Total frequencies per feature and class.
        class_totals (collections.Counter): Total frequencies per class.

    Example:

        >>> import math
        >>> from creme import compose
        >>> from creme import feature_extraction
        >>> from creme import naive_bayes

        >>> docs = [
        ...     ('Chinese Beijing Chinese', 'yes'),
        ...     ('Chinese Chinese Shanghai', 'yes'),
        ...     ('Chinese Macao', 'yes'),
        ...     ('Tokyo Japan Chinese', 'no')
        ... ]
        >>> model = compose.Pipeline(
        ...     ('tokenize', feature_extraction.BagOfWords(lowercase=False)),
        ...     ('nb', naive_bayes.MultinomialNB(alpha=1))
        ... )
        >>> for sentence, label in docs:
        ...     model = model.fit_one(sentence, label)

        >>> model['nb'].p_class('yes')
        0.75
        >>> model['nb'].p_class('no')
        0.25

        >>> cp = model['nb'].p_feature_given_class

        >>> cp('Chinese', 'yes') == (5 + 1) / (8 + 6)
        True

        >>> cp('Tokyo', 'yes') == (0 + 1) / (8 + 6)
        True
        >>> cp('Japan', 'yes') == (0 + 1) / (8 + 6)
        True

        >>> cp('Chinese', 'no') == (1 + 1) / (3 + 6)
        True

        >>> cp('Tokyo', 'no') == (1 + 1) / (3 + 6)
        True
        >>> cp('Japan', 'no') == (1 + 1) / (3 + 6)
        True

        >>> new_text = 'Chinese Chinese Chinese Tokyo Japan'
        >>> tokens = model['tokenize'].transform_one(new_text)
        >>> jlh = model['nb'].joint_log_likelihood(tokens)
        >>> math.exp(jlh['yes'])
        0.000301
        >>> math.exp(jlh['no'])
        0.000135
        >>> model.predict_one(new_text)
        'yes'

        >>> new_unseen_text = 'Taiwanese Taipei'
        >>> tokens = model['tokenize'].transform_one(new_unseen_text)
        >>> # P(Taiwanese|yes)
        >>> #   = (N_Taiwanese_yes + 1) / (N_yes + N_terms)
        >>> cp('Taiwanese', 'yes') == cp('Taipei', 'yes') == (0 + 1) / (8 + 6)
        True
        >>> cp('Taiwanese', 'no') == cp('Taipei', 'no') == (0 + 1) / (3 + 6)
        True

        >>> # P(yes|Taiwanese Taipei)
        >>> #   ∝ P(Taiwanese|yes) * P(Taipei|yes) * P(yes)
        >>> posterior_yes_given_new_text = (0 + 1) / (8 + 6) * (0 + 1) / (8 + 6) * 0.75
        >>> jlh = model['nb'].joint_log_likelihood(tokens)
        >>> jlh['yes'] == math.log(posterior_yes_given_new_text)
        True

        >>> model.predict_one(new_unseen_text)
        'yes'

    References:
        1. [Naive Bayes text classification](https://nlp.stanford.edu/IR-book/html/htmledition/naive-bayes-text-classification-1.html)

    """

    def __init__(self, alpha=1.):
        self.alpha = alpha
        self.class_counts = collections.Counter()
        self.feature_counts = collections.defaultdict(collections.Counter)
        self.class_totals = collections.Counter()

    @property
    def _more_tags(self):
        return {'requires_positive_data': True}

    def fit_one(self, x, y):
        self.class_counts.update((y,))

        for f, frequency in x.items():
            self.feature_counts[f].update({y: frequency})
            self.class_totals.update({y: frequency})

        return self

    @property
    def classes_(self):
        return list(self.class_counts.keys())

    @property
    def n_terms(self):
        return len(self.feature_counts)

    def p_feature_given_class(self, f, c):
        num = self.feature_counts.get(f, {}).get(c, 0.) + self.alpha
        den = self.class_totals[c] + self.alpha * self.n_terms
        return num / den

    def p_class(self, c):
        return self.class_counts[c] / sum(self.class_counts.values())

    def joint_log_likelihood(self, x):
        return {
            c: math.log(self.p_class(c)) + sum(
                frequency * math.log(self.p_feature_given_class(f, c))
                for f, frequency in x.items()
            )
            for c in self.classes_
        }



cimport creme.stats.base

cdef class Mean(creme.stats.base.Univariate):
    cdef readonly double mean
    cdef readonly double n

    cpdef double get(self)
    cpdef Mean update(self, double x, double w=*)
    cpdef Mean revert(self, double x, double w=*)



import math

from creme import utils

from . import base


class Min(base.Univariate):
    """Running min.

    Attributes:
        min (float): The current min.

    """

    def __init__(self):
        self.min = math.inf

    def update(self, x):
        if x < self.min:
            self.min = x
        return self

    def get(self):
        return self.min


class RollingMin(base.RollingUnivariate, utils.SortedWindow):
    """Running min over a window.

    Parameters:
        window_size: Size of the rolling window.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 2, 1]
        >>> rolling_min = stats.RollingMin(2)
        >>> for x in X:
        ...     print(rolling_min.update(x).get())
        1
        -4
        -4
        -2
        -2
        1

    """

    def __init__(self, window_size: int):
        super().__init__(size=window_size)

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        self.append(x)
        return self

    def get(self):
        return self[0]



import functools
import gzip
import io
import os
import zipfile


def open_filepath(filepath_or_buffer, compression):

    # Determine the compression from the file extension if "infer" has been specified
    if compression == 'infer':
        _, ext = os.path.splitext(filepath_or_buffer)
        compression = {'.gz': 'gzip', '.zip': 'zip'}.get(ext)

    def open_zipfile(path):

        with zipfile.ZipFile(path, 'r') as zf:
            f = zf.open(zf.namelist()[0], 'r')
            f = io.TextIOWrapper(f)
            return f

    # Determine the file opening method from the compression
    open_func = {
        None: open,
        'gzip': functools.partial(gzip.open, mode='rt'),
        'zip': open_zipfile
    }[compression]

    # Open the file using the opening method
    return open_func(filepath_or_buffer)



import math

from . import knn


def test_find_nearest_manhattan():

    X = [
        {'x': 0, 'y': 0},
        {'x': 0, 'y': 1},
        {'x': 1, 'y': 0},
        {'x': 1, 'y': 1},
    ]

    nn = knn.NearestNeighbours(window_size=len(X), p=1)

    for x in X:
        nn.update(x, None)

    neighbors = nn.find_nearest(x={'x': 0, 'y': .4}, k=2)

    assert neighbors[0][0] == X[0]
    assert math.isclose(neighbors[0][2], .4)
    assert neighbors[1][0] == X[1]
    assert math.isclose(neighbors[1][2], .6)


def test_find_nearest_euclidean():

    X = [
        {'x': 0, 'y': 0},
        {'x': 0, 'y': 1},
        {'x': 1, 'y': 0},
        {'x': 1, 'y': 1},
    ]

    nn = knn.NearestNeighbours(window_size=len(X), p=2)

    for x in X:
        nn.update(x, None)

    neighbors = nn.find_nearest(x={'x': 0, 'y': .4}, k=2)

    assert neighbors[0][0] == X[0]
    assert math.isclose(neighbors[0][2], .16)
    assert neighbors[1][0] == X[1]
    assert math.isclose(neighbors[1][2], .36)


def test_regression_predict():

    X_y = [
        ({'x': 0, 'y': 0}, 3),
        ({'x': 0, 'y': 1}, 4),
        ({'x': 1, 'y': 0}, 11),
        ({'x': 1, 'y': 1}, 12),
    ]

    model = knn.KNeighborsRegressor(n_neighbors=2, window_size=len(X_y), p=1, weighted=True)

    for x, y in X_y:
        model._nn.update(x, y)

    # X_y[0] and X_y[1] should be the closest
    y_pred = model.predict_one(x={'x': 0, 'y': .4})
    expected = (
        (X_y[0][1] / .4 + X_y[1][1] / .6) /
        (1 / .4 + 1 / .6)
    )
    assert math.isclose(y_pred, expected)

    # X_y[0] is the closest
    y_pred = model.predict_one(x={'x': 0, 'y': 0})
    expected = 3
    assert math.isclose(y_pred, expected)



cimport creme.stats.base

from . import base
from . import summing


cdef class Mean(creme.stats.base.Univariate):
    """Running mean.

    Parameters:
        mean: Initial mean.
        n: Initial sum of weights.

    Attributes:
        mean (float): The current value of the mean.
        n (float): The current sum of weights. If each passed weight was 1, then this is equal to
            the number of seen observations.

    Example:

        >>> from creme import stats

        >>> X = [-5, -3, -1, 1, 3, 5]
        >>> mean = stats.Mean()
        >>> for x in X:
        ...     print(mean.update(x).get())
        -5.0
        -4.0
        -3.0
        -2.0
        -1.0
        0.0

    References:
        1. [West, D. H. D. (1979). Updating mean and variance estimates: An improved method. Communications of the ACM, 22(9), 532-535.](https://people.xiph.org/~tterribe/tmp/homs/West79-_Updating_Mean_and_Variance_Estimates-_An_Improved_Method.pdf)
        2. [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf)

    """

    cpdef Mean update(self, double x, double w=1.):
        self.n += w
        if self.n > 0:
            self.mean += w * (x - self.mean) / self.n
        return self

    cpdef Mean revert(self, double x, double w=1.):
        self.n -= w
        if self.n < 0:
            raise ValueError('Cannot go below 0')
        elif self.n == 0:
            self.mean = 0.
        else:
            self.mean -= w * (x - self.mean) / self.n
        return self

    cpdef double get(self):
        return self.mean


class RollingMean(summing.RollingSum):
    """Running average over a window.

    Parameters:
        window_size: Size of the rolling window.

    Example:

        >>> import creme

        >>> X = [1, 2, 3, 4, 5, 6]

        >>> rmean = creme.stats.RollingMean(window_size=2)
        >>> for x in X:
        ...     print(rmean.update(x).get())
        1.0
        1.5
        2.5
        3.5
        4.5
        5.5

        >>> rmean = creme.stats.RollingMean(window_size=3)
        >>> for x in X:
        ...     print(rmean.update(x).get())
        1.0
        1.5
        2.0
        3.0
        4.0
        5.0

    """

    def get(self):
        return super().get() / len(self) if len(self) > 0 else 0


class BayesianMean(base.Univariate):
    """Estimates a mean using outside information.

    Parameters:
        prior
        prior_weight

    References:
        1. [Additive smoothing](https://www.wikiwand.com/en/Additive_smoothing)
        2. [Bayesian average](https://www.wikiwand.com/en/Bayesian_average)
        3. [Practical example of Bayes estimators](https://www.wikiwand.com/en/Bayes_estimator#/Practical_example_of_Bayes_estimators)

    """

    def __init__(self, prior: float, prior_weight: float):
        self.prior = prior
        self.prior_weight = prior_weight
        self.mean = Mean()

    @property
    def name(self):
        return 'bayes_mean'

    def update(self, x):
        self.mean.update(x)
        return self

    def get(self):

        # Uses the notation from https://www.wikiwand.com/en/Bayes_estimator#/Practical_example_of_Bayes_estimators
        R = self.mean.get()
        v = self.mean.n
        m = self.prior_weight
        C = self.prior

        return (R * v + C * m) / (v + m)



import collections

import numpy as np

from . import base


__all__ = ['FTRLProximal']


class FTRLProximal(base.Optimizer):
    """FTRL-Proximal optimizer.

    Parameters:
        alpha
        beta
        l1
        l2

    Attributes:
        z (collections.defaultdict)
        n (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.FTRLProximal()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.876588

    References:
        1. [McMahan, H.B., Holt, G., Sculley, D., Young, M., Ebner, D., Grady, J., Nie, L., Phillips, T., Davydov, E., Golovin, D. and Chikkerur, S., 2013, August. Ad click prediction: a view from the trenches. In Proceedings of the 19th ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 1222-1230)](https://static.googleusercontent.com/media/research.google.com/en//pubs/archive/41159.pdf)
        2. [Tensorflow's `FtrlOptimizer`](https://www.tensorflow.org/api_docs/python/tf/train/FtrlOptimizer)

    """

    def __init__(self, alpha=.05, beta=1., l1=0., l2=1.):
        self.alpha = alpha
        self.beta = beta
        self.l1 = l1
        self.l2 = l2
        self.z = collections.defaultdict(float)
        self.n = collections.defaultdict(float)
        self.n_iterations = 0

    def _update_after_pred(self, w, g):

        alpha = self.alpha
        beta = self.beta
        l1 = self.l1
        l2 = self.l2
        z = self.z
        n = self.n

        for i in g:
            if abs(z[i]) > l1:
                w[i] = -((beta + n[i] ** 0.5) / alpha + l2) ** -1 * (z[i] - np.sign(z[i]) * l1)

        for i, gi in g.items():
            s = ((self.n[i] + gi ** 2) ** 0.5 - self.n[i] ** 0.5) / self.alpha
            self.z[i] += gi - s * w.get(i, 0)
            self.n[i] += gi ** 2

        return w



"""Mathematical utility functions (intended for internal purposes).

A lot of this is experimental and has a high probability of changing in the future.

"""
import functools
import itertools
import math
import operator

import numpy as np


__all__ = [
    'chain_dot',
    'clamp',
    'dot',
    'dotvecmat',
    'matmul2d',
    'minkowski_distance',
    'norm',
    'outer',
    'prod',
    'sigmoid',
    'sign',
    'sherman_morrison',
    'softmax'
]


def sherman_morrison(A_inv: dict, u: dict, v: dict) -> dict:
    """Sherman–Morrison formula.

    This modifies `A_inv` inplace.

    Parameters:
        A_inv
        u
        v

    References:
        1. [Wikipedia article on the Sherman-Morrison formula](https://www.wikiwand.com/en/Sherman%E2%80%93Morrison_formula)s

    Example:

        >>> import pprint
        >>> from creme import utils

        >>> A_inv = {
        ...     (0, 0): 0.2,
        ...     (1, 1): 1,
        ...     (2, 2): 1
        ... }
        >>> u = {0: 1, 1: 2, 2: 3}
        >>> v = {0: 4}

        >>> inv = sherman_morrison(A_inv, u, v)
        >>> pprint.pprint(inv)
        {(0, 0): 0.111111,
            (1, 0): -0.888888,
            (1, 1): 1,
            (2, 0): -1.333333,
            (2, 2): 1}

    """

    den = 1 + dot(dotvecmat(u, A_inv), v)

    for k, v in matmul2d(matmul2d(A_inv, outer(u, v)), A_inv).items():
        A_inv[k] = A_inv.get(k, 0) - v / den

    return A_inv


def dotvecmat(x, A):

    C = {}

    for (i, xi), ((j, k), ai) in itertools.product(x.items(), A.items()):

        if i != j:
            continue

        C[k] = C.get(j, 0.) + xi * ai

    return C


def matmul2d(A, B):
    """Multiplication for 2D matrices.

    Parameters:
        A
        B

    Example:

        >>> import pprint
        >>> from creme import utils

        >>> A = {
        ...     (0, 0): 2, (0, 1): 0, (0, 2): 4,
        ...     (1, 0): 5, (1, 1): 6, (1, 2): 0
        ... }

        >>> B = {
        ...     (0, 0): 1, (0, 1): 1, (0, 2): 0, (0, 3): 0,
        ...     (1, 0): 2, (1, 1): 0, (1, 2): 1, (1, 3): 3,
        ...     (2, 0): 4, (2, 1): 0, (2, 2): 0, (2, 3): 0
        ... }

        >>> C = matmul2d(A, B)
        >>> pprint.pprint(C)
        {(0, 0): 18.0,
            (0, 1): 2.0,
            (0, 2): 0.0,
            (0, 3): 0.0,
            (1, 0): 17.0,
            (1, 1): 5.0,
            (1, 2): 6.0,
            (1, 3): 18.0}

    """
    C = {}

    for ((i, k1), x), ((k2, j), y) in itertools.product(A.items(), B.items()):
        if k1 != k2:
            continue
        C[i, j] = C.get((i, j), 0.) + x * y

    return C


def outer(u: dict, v: dict) -> dict:
    """Outer-product between two vectors.

    Parameters:
        u
        v

    Example:

        >>> import pprint
        >>> from creme import utils

        >>> u = dict(enumerate((1, 2, 3)))
        >>> v = dict(enumerate((2, 4, 8)))

        >>> uTv = utils.math.outer(u, v)
        >>> pprint.pprint(uTv)
        {(0, 0): 2,
            (0, 1): 4,
            (0, 2): 8,
            (1, 0): 4,
            (1, 1): 8,
            (1, 2): 16,
            (2, 0): 6,
            (2, 1): 12,
            (2, 2): 24}

    """
    return {
        (ki, kj): vi * vj
        for (ki, vi), (kj, vj) in itertools.product(u.items(), v.items())
    }


def minkowski_distance(a: dict, b: dict, p: int):
    """Minkowski distance.

    Parameters:
        a
        b
        p: Parameter for the Minkowski distance. When `p=1`, this is equivalent to using the
            Manhattan distance. When `p=2`, this is equivalent to using the Euclidean distance.

    """
    return sum((abs(a.get(k, 0.) - b.get(k, 0.))) ** p for k in set([*a.keys(), *b.keys()]))


def softmax(y_pred):
    """Normalizes a dictionary of predicted probabilities, in-place."""

    if not y_pred:
        return y_pred

    maximum = max(y_pred.values())
    total = 0.

    for c, p in y_pred.items():
        y_pred[c] = math.exp(p - maximum)
        total += y_pred[c]

    for c in y_pred:
        y_pred[c] /= total

    return y_pred


def prod(iterable):
    return functools.reduce(operator.mul, iterable, 1)


def dot(x: dict, y: dict):
    """Returns the dot product of two vectors represented as dicts.

    Example:

        >>> from creme import utils

        >>> x = {'x0': 1, 'x1': 2}
        >>> y = {'x1': 21, 'x2': 3}

        >>> utils.math.dot(x, y)
        42

    """

    if len(x) < len(y):
        return sum(xi * y[i] for i, xi in x.items() if i in y)
    return sum(x[i] * yi for i, yi in y.items() if i in x)


def chain_dot(*xs):
    """Returns the dot product of multiple vectors represented as dicts.

    Example:

        ::

            >>> from creme import utils

            >>> x = {'x0': 1, 'x1': 2, 'x2': 1}
            >>> y = {'x1': 21, 'x2': 3}
            >>> z = {'x1': 2, 'x2': 1 / 3}

            >>> utils.math.chain_dot(x, y, z)
            85.0

    """
    keys = min(xs, key=len)
    return sum(prod(x.get(i, 0) for x in xs) for i in keys)


def sigmoid(x: float):
    if x < -30:
        return 0
    if x > 30:
        return 1
    return 1 / (1 + math.exp(-x))


def clamp(x: float, minimum=0., maximum=1.):
    return max(min(x, maximum), minimum)


def norm(x, order=None):
    return np.linalg.norm(list(x.values()), ord=order)

def sign(x: float):
    return -1 if x < 0 else (1 if x > 0 else 0)



import collections
import typing

from creme import base


__all__ = ['OneHotEncoder']


class OneHotEncoder(base.Transformer):
    """One-hot encoding.

    This transformer will encode every feature it is provided it with. You can apply it to a
    subset of features by composing it with `compose.Select` or `compose.SelectType`.

    Parameters:
        sparse: Whether or not 0s should be made explicit or not.

    Example:

        Let us first create an example dataset.

        >>> from pprint import pprint
        >>> import random
        >>> import string

        >>> random.seed(42)
        >>> alphabet = list(string.ascii_lowercase)
        >>> X = [
        ...     {
        ...         'c1': random.choice(alphabet),
        ...         'c2': random.choice(alphabet),
        ...     }
        ...     for _ in range(4)
        ... ]
        >>> pprint(X)
        [{'c1': 'u', 'c2': 'd'},
         {'c1': 'a', 'c2': 'x'},
         {'c1': 'i', 'c2': 'h'},
         {'c1': 'h', 'c2': 'e'}]

        We can now apply one-hot encoding. All the provided are one-hot encoded, there is therefore
        no need to specify which features to encode.

        >>> import creme.preprocessing

        >>> oh = creme.preprocessing.OneHotEncoder(sparse=True)
        >>> for x in X:
        ...     oh = oh.fit_one(x)
        ...     pprint(oh.transform_one(x))
        {'c1_u': 1, 'c2_d': 1}
        {'c1_a': 1, 'c2_x': 1}
        {'c1_i': 1, 'c2_h': 1}
        {'c1_h': 1, 'c2_e': 1}

        The `sparse` parameter can be set to `False` in order to include the values that are not
        present in the output.

        >>> oh = creme.preprocessing.OneHotEncoder(sparse=False)
        >>> for x in X[:2]:
        ...     oh = oh.fit_one(x)
        ...     pprint(oh.transform_one(x))
        {'c1_u': 1, 'c2_d': 1}
        {'c1_a': 1, 'c1_u': 0, 'c2_d': 0, 'c2_x': 1}

        A subset of the features can be one-hot encoded by using an instance of `compose.Select`.

        >>> from creme import compose

        >>> pp = compose.Select('c1') | creme.preprocessing.OneHotEncoder()

        >>> for x in X:
        ...     pp = pp.fit_one(x)
        ...     pprint(pp.transform_one(x))
        {'c1_u': 1}
        {'c1_a': 1, 'c1_u': 0}
        {'c1_a': 0, 'c1_i': 1, 'c1_u': 0}
        {'c1_a': 0, 'c1_h': 1, 'c1_i': 0, 'c1_u': 0}

        You can preserve the `c2` feature by using a union:

        >>> pp = compose.Select('c1') | creme.preprocessing.OneHotEncoder()
        >>> pp += compose.Select('c2')

        >>> for x in X:
        ...     pp = pp.fit_one(x)
        ...     pprint(pp.transform_one(x))
        {'c1_u': 1, 'c2': 'd'}
        {'c1_a': 1, 'c1_u': 0, 'c2': 'x'}
        {'c1_a': 0, 'c1_i': 1, 'c1_u': 0, 'c2': 'h'}
        {'c1_a': 0, 'c1_h': 1, 'c1_i': 0, 'c1_u': 0, 'c2': 'e'}

    """

    def __init__(self, sparse=False):
        self.sparse = sparse
        self.values = collections.defaultdict(set)

    def fit_one(self, x):
        for i, xi in x.items():
            self.values[i].add(xi)
        return self

    def transform_one(self, x, y=None):
        oh = {}

        # Add 0s
        if not self.sparse:
            oh = {f'{i}_{v}': 0 for i, values in self.values.items() for v in values}

        # Add 1s
        for i, xi in x.items():
            oh[f'{i}_{xi}'] = 1

        return oh



import abc


cdef class Statistic:
    """A statistic."""

    # Define the format specification used for string representation.
    _fmt = ',.6f'  # Use commas to separate big numbers and show 6 decimals

    cpdef double get(self):
        """Returns the current value of the statistic."""
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.get():{self._fmt}}'.rstrip('0')


cdef class Univariate(Statistic):
    """A univariate statistic measures a property of a variable."""

    cpdef Univariate update(self, double x):
        """Updates and returns the called instance."""
        raise NotImplementedError

    cpdef Univariate revert(self, double x):
        """Reverts and returns the called instance."""
        raise NotImplementedError

    @property
    def name(self):
        return self.__class__.__name__.lower()

    def __or__(self, other):
        from .link import Link
        return Link(left=self, right=other)


class RollingUnivariate(Univariate):
    """A rolling univariate statistic measures a property of a variable over a window."""

    @abc.abstractproperty
    def window_size(self):
        pass

    @property
    def name(self):
        return f'{super().name}_{self.window_size}'


cdef class Bivariate(Statistic):
    """A bivariate statistic measures a relationship between two variables."""

    cpdef Bivariate update(self, double x, double y):
        """Updates and returns the called instance."""
        raise NotImplementedError



print("import: 'creme'")
import creme

print("import: 'creme.anomaly'")
import creme.anomaly

print("import: 'creme.cluster'")
import creme.cluster

print("import: 'creme.compat'")
import creme.compat

print("import: 'creme.compose'")
import creme.compose

print("import: 'creme.decomposition'")
import creme.decomposition

print("import: 'creme.ensemble'")
import creme.ensemble

print("import: 'creme.feature_extraction'")
import creme.feature_extraction

print("import: 'creme.feature_selection'")
import creme.feature_selection

print("import: 'creme.impute'")
import creme.impute

print("import: 'creme.linear_model'")
import creme.linear_model

print("import: 'creme.metrics'")
import creme.metrics

print("import: 'creme.multiclass'")
import creme.multiclass

print("import: 'creme.multioutput'")
import creme.multioutput

print("import: 'creme.naive_bayes'")
import creme.naive_bayes

print("import: 'creme.neighbors'")
import creme.neighbors

print("import: 'creme.optim'")
import creme.optim

print("import: 'creme.optim.losses'")
import creme.optim.losses

print("import: 'creme.preprocessing'")
import creme.preprocessing

print("import: 'creme.proba'")
import creme.proba

print("import: 'creme.reco'")
import creme.reco

print("import: 'creme.stats'")
import creme.stats

print("import: 'creme.tree'")
import creme.tree

print("import: 'creme.utils'")
import creme.utils




import math
import random

from .. import base


class Friedman(base.SyntheticDataset):
    """Friedman synthetic dataset.

    Each observation is composed of 10 features. Each feature value is sampled uniformly in [0, 1].
    The target is defined by the following function:

    $$y = 10 sin(\\pi x_0 x_1) + 20 (x_2 - .5)^2 + 10 x_3 + 5 x_4$$

    Therefore, only the first 5 features are relevant.

    Parameters:
        seed: Random seed number used for reproducibility.

    Example:

        >>> from creme import datasets

        >>> dataset = datasets.synth.Friedman(seed=42)

        >>> for x, y in dataset.take(5):
        ...     print(list(x.values()), y)
        [0.63, 0.02, 0.27, 0.22, 0.73, 0.67, 0.89, 0.08, 0.42, 0.02] 7.42
        [0.21, 0.50, 0.02, 0.19, 0.64, 0.54, 0.22, 0.58, 0.80, 0.00] 13.12
        [0.80, 0.69, 0.34, 0.15, 0.95, 0.33, 0.09, 0.09, 0.84, 0.60] 16.65
        [0.80, 0.72, 0.53, 0.97, 0.37, 0.55, 0.82, 0.61, 0.86, 0.57] 21.26
        [0.70, 0.04, 0.22, 0.28, 0.07, 0.23, 0.10, 0.27, 0.63, 0.36] 5.78

    References:
        1. [Friedman, J.H., 1991. Multivariate adaptive regression splines. The annals of statistics, pp.1-67.](https://projecteuclid.org/euclid.aos/1176347963)

    """

    def __init__(self, seed: int = None):
        self._rng = random.Random(seed)

    def __iter__(self):

        while True:

            x = {i: self._rng.uniform(a=0, b=1) for i in range(10)}
            y = 10 * math.sin(math.pi * x[0] * x[1]) + 20 * (x[2] - .5) ** 2 + 10 * x[3] + 5 * x[4]

            yield x, y



import collections
import operator

from creme import base
from creme import utils


__all__ = ['KNeighborsRegressor', 'KNeighborsClassifier']


class NearestNeighbours:

    def __init__(self, window_size, p):
        self.window_size = window_size
        self.p = p
        self.window = collections.deque(maxlen=window_size)

    def update(self, x, y):
        self.window.append((x, y))
        return self

    def find_nearest(self, x, k):
        """Returns the `k` closest points to `x`, along with their distances."""

        # Compute the distances to each point in the window
        points = ((*p, utils.math.minkowski_distance(a=x, b=p[0], p=self.p)) for p in self.window)

        # Return the k closest points
        return sorted(points, key=operator.itemgetter(2))[:k]


class KNeighborsRegressor(base.Regressor):
    """K-Nearest Neighbors (KNN) for regression.

    This works by storing a buffer with the `window_size` most recent observations. A brute-force
    search is used to find the `n_neighbors` nearest observations in the buffer to make a
    prediction.

    Parameters:
        n_neighbors: Number of neighbors to use.
        window_size: Size of the sliding window use to search neighbors with.
        p: Power parameter for the Minkowski metric. When `p=1`, this corresponds to the
            Manhattan distance, while `p=2` corresponds to the Euclidean distance.
        weighted: Whether to weight the contribution of each neighbor by it's inverse distance or
            not.

    Example:

        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import neighbors
        >>> from creme import preprocessing

        >>> X_y = datasets.TrumpApproval()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     neighbors.KNeighborsRegressor()
        ... )

        >>> metric = metrics.MAE()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MAE: 0.335753

    """

    def __init__(self, n_neighbors=5, window_size=50, p=2, weighted=True):
        self.n_neighbors = n_neighbors
        self.window_size = window_size
        self.weighted = weighted
        self.p = p
        self._nn = NearestNeighbours(window_size=window_size, p=p)

    def fit_one(self, x, y):
        self._nn.update(x, y)
        return self

    def predict_one(self, x):

        nearest = self._nn.find_nearest(x=x, k=self.n_neighbors)

        if not nearest:
            return 0.

        # If the closest neighbor has a distance of 0, then return it's output
        if nearest[0][2] == 0:
            return nearest[0][1]

        # Weighted average
        if self.weighted:
            return (
                sum(y / d for _, y, d in nearest) /
                sum(1 / d for _, _, d in nearest)
            )

        # Uniform average
        return sum(y for _, y, _ in nearest) / self.n_neighbors


class KNeighborsClassifier(base.MultiClassifier):
    """K-Nearest Neighbors (KNN) for classification.

    This works by storing a buffer with the `window_size` most recent observations. A brute-force
    search is used to find the `n_neighbors` nearest observations in the buffer to make a
    prediction.

    Parameters:
        n_neighbors: Number of neighbors to use.
        window_size: Size of the sliding window use to search neighbors with.
        p: Power parameter for the Minkowski metric. When `p=1`, this corresponds to the
            Manhattan distance, while `p=2` corresponds to the Euclidean distance.
        weighted: Whether to weight the contribution of each neighbor by it's inverse
            distance or not.

    Example:

        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import neighbors
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     neighbors.KNeighborsClassifier()
        ... )

        >>> metric = metrics.Accuracy()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        Accuracy: 84.55%

    """

    def __init__(self, n_neighbors=5, window_size=50, p=2, weighted=True):
        self.n_neighbors = n_neighbors
        self.window_size = window_size
        self.weighted = weighted
        self.p = p
        self.classes = set()
        self._nn = NearestNeighbours(window_size=window_size, p=p)

    def fit_one(self, x, y):
        self.classes.add(y)
        self._nn.update(x, y)
        return self

    def predict_proba_one(self, x):

        nearest = self._nn.find_nearest(x=x, k=self.n_neighbors)

        y_pred = {c: 0. for c in self.classes}

        if not nearest:
            return y_pred

        # If the closest neighbor has a distance of 0, then return it's output
        if nearest[0][2] == 0:
            y_pred[nearest[0][1]] = 1.
            return y_pred

        # Weighted votes
        if self.weighted:
            for _, y, d in nearest:
                y_pred[y] += 1. / d

        # Uniform votes
        else:
            for _, y, _ in nearest:
                y_pred[y] += 1.

        # Normalize votes into real [0, 1] probabilities
        return utils.math.softmax(y_pred)



from creme import base


__all__ = ['PreviousImputer']


class PreviousImputer(base.Transformer):
    """Imputes missing values by using the most recent value.

    Example:

        >>> from creme import impute

        >>> imputer = impute.PreviousImputer()

        >>> imputer = imputer.fit_one({'x': 1, 'y': 2})
        >>> imputer.transform_one({'y': None})
        {'y': 2}

        >>> imputer.transform_one({'x': None})
        {'x': 1}

    """

    def __init__(self):
        self._latest = {}

    def fit_one(self, x):

        for i, v in x.items():
            if v is not None:
                self._latest[i] = v

        return self

    def transform_one(self, x):

        for i, v in x.items():
            if v is None:
                x[i] = self._latest.get(i)

        return x



import math

from . import base


__all__ = ['LogLoss']


class LogLoss(base.MeanMetric, base.BinaryMetric):
    """Binary logarithmic loss.

    Example:

        >>> from creme import metrics

        >>> y_true = [True, False, False, True]
        >>> y_pred = [0.9,  0.1,   0.2,   0.65]

        >>> metric = metrics.LogLoss()
        >>> for yt, yp in zip(y_true, y_pred):
        ...     metric = metric.update(yt, yp)
        ...     print(metric.get())
        0.105360
        0.105360
        0.144621
        0.216161

        >>> metric
        LogLoss: 0.216162

    """

    @property
    def bigger_is_better(self):
        return False

    @property
    def requires_labels(self):
        return False

    def _eval(self, y_true, y_pred):
        p_true = y_pred.get(True, 0.) if isinstance(y_pred, dict) else y_pred
        p_true = self._clamp_proba(p_true)
        if y_true:
            return -math.log(p_true)
        return -math.log(1 - p_true)



from .. import base


__all__ = ['Renamer']


class Renamer(base.Transformer):
    """Renames keys based on given parameters.

    Parameters:
        prefix
        suffix

    Example:

        >>> from creme import compose

        >>> x = {'a': 42, 'b': 12}
        >>> compose.Renamer(prefix='prefix_', suffix='_suffix').transform_one(x)
        {'prefix_a_suffix': 42, 'prefix_b_suffix': 12}

    """

    def __init__(self, prefix=None, suffix=None):
        self.prefix = prefix or ''
        self.suffix = suffix or ''

    def _rename(self, s):
        return self.prefix + s + self.suffix

    def transform_one(self, x):
        return {self._rename(i): xi for i, xi in x.items()}



import collections
import functools
import itertools

from .. import base
from .. import optim
from .. import utils

from .base import BaseFM


__all__ = [
    'HOFMClassifier',
    'HOFMRegressor'
]


class HOFM(BaseFM):
    """Higher-Order Factorization Machines.

    Parameters:
        degree (int): Polynomial degree or model order.
        n_factors (int): Dimensionality of the factorization or number of latent factors.
        weight_optimizer (optim.Optimizer): The sequential optimizer used for updating the feature
            weights. Note that the intercept is handled separately.
        latent_optimizer (optim.Optimizer): The sequential optimizer used for updating the latent
            factors.
        loss (optim.Loss): The loss function to optimize for.
        sample_normalization (bool): Whether to divide each element of ``x`` by ``x`` L2-norm.
            Defaults to False.
        l1_weight (float): Amount of L1 regularization used to push weights towards 0.
        l2_weight (float): Amount of L2 regularization used to push weights towards 0.
        l1_latent (float): Amount of L1 regularization used to push latent weights towards 0.
        l2_latent (float): Amount of L2 regularization used to push latent weights towards 0.
        intercept (float or `stats.Univariate` instance): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        weight_initializer (optim.initializers.Initializer): Weights initialization scheme.
        latent_initializer (optim.initializers.Initializer): Latent factors initialization scheme.
        clip_gradient (float): Clips the absolute value of each gradient value.
        seed (int): Randomization seed used for reproducibility.

    Attributes:
        weights (collections.defaultdict): The current weights assigned to the features.
        latents (collections.defaultdict): The current latent weights assigned to the features.

    """

    def __init__(self, degree, n_factors, weight_optimizer, latent_optimizer, loss,
                 sample_normalization, l1_weight, l2_weight, l1_latent, l2_latent, intercept,
                 intercept_lr, weight_initializer, latent_initializer, clip_gradient, seed):
        super().__init__(
            n_factors=n_factors,
            weight_optimizer=weight_optimizer,
            latent_optimizer=latent_optimizer,
            loss=loss,
            sample_normalization=sample_normalization,
            l1_weight=l1_weight,
            l2_weight=l2_weight,
            l1_latent=l1_latent,
            l2_latent=l2_latent,
            intercept=intercept,
            intercept_lr=intercept_lr,
            weight_initializer=weight_initializer,
            latent_initializer=latent_initializer,
            clip_gradient=clip_gradient,
            seed=seed
        )
        self.degree = degree

    def _init_latents(self):
        random_latents = functools.partial(
            self.latent_initializer,
            shape=self.n_factors
        )
        order_latents_dict = functools.partial(
            collections.defaultdict, random_latents
        )
        return collections.defaultdict(order_latents_dict)

    def _calculate_interactions(self, x):
        """Calculates greater than unary interactions."""
        return sum(
            self._calculate_interaction(x, l, combination)
            for l in range(2, self.degree + 1)
            for combination in itertools.combinations(x.keys(), l)
        )

    def _calculate_interaction(self, x, l, combination):
        feature_product = functools.reduce(lambda x, y: x * y, (x[j] for j in combination))
        latent_scalar_product = sum(
            functools.reduce(lambda x, y: x * y, (self.latents[j][l][f] for j in combination))
            for f in range(self.n_factors)
        )
        return feature_product * latent_scalar_product

    def _calculate_weights_gradients(self, x, g_loss):

        # For notational convenience
        w, l1, l2, sign = self.weights, self.l1_weight, self.l2_weight, utils.math.sign

        return {
            j: g_loss * xj + l1 * sign(w[j]) + l2 * w[j]
            for j, xj in x.items()
        }

    def _update_latents(self, x, g_loss):

        # For notational convenience
        v, l1, l2, sign = self.latents, self.l1_latent, self.l2_latent, utils.math.sign

        # Calculate each latent factor gradient before updating any
        gradients = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(float)
            )
        )

        for l in range(2, self.degree + 1):

            for combination in itertools.combinations(x.keys(), l):
                feature_product = functools.reduce(lambda x, y: x * y, (x[j] for j in combination))

                for f in range(self.n_factors):
                    latent_product = functools.reduce(lambda x, y: x * y, (v[j][l][f] for j in combination))

                    for j in combination:
                        gradients[j][l][f] += feature_product * latent_product / v[j][l][f]

        # Finally update the latent weights
        for j in x.keys():
            for l in range(2, self.degree + 1):
                self.latents[j][l] = self.latent_optimizer.update_after_pred(
                    w=v[j][l],
                    g={
                        f: g_loss * gradients[j][l][f] + l1 * sign(v[j][l][f]) + 2 * l2 * v[j][l][f]
                        for f in range(self.n_factors)
                    }
                )


class HOFMRegressor(HOFM, base.Regressor):
    """Higher-Order Factorization Machines Regressor.

    Parameters:
        degree (int): Polynomial degree or model order.
        n_factors (int): Dimensionality of the factorization or number of latent factors.
        weight_optimizer (optim.Optimizer): The sequential optimizer used for updating the feature
            weights. Note that the intercept is handled separately.
        latent_optimizer (optim.Optimizer): The sequential optimizer used for updating the latent
            factors.
        loss (optim.Loss): The loss function to optimize for.
        sample_normalization (bool): Whether to divide each element of ``x`` by ``x`` L2-norm.
            Defaults to False.
        l1_weight (float): Amount of L1 regularization used to push weights towards 0.
        l2_weight (float): Amount of L2 regularization used to push weights towards 0.
        l1_latent (float): Amount of L1 regularization used to push latent weights towards 0.
        l2_latent (float): Amount of L2 regularization used to push latent weights towards 0.
        intercept (float or `stats.Univariate` instance): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        weight_initializer (optim.initializers.Initializer): Weights initialization scheme. Defaults
            to ``optim.initializers.Zeros()``.
        latent_initializer (optim.initializers.Initializer): Latent factors initialization scheme.
            Defaults to
            ``optim.initializers.Normal(mu=.0, sigma=.1, random_state=self.random_state)``.
        clip_gradient (float): Clips the absolute value of each gradient value.
        seed (int): Randomization seed used for reproducibility.

    Attributes:
        weights (collections.defaultdict): The current weights assigned to the features.
        latents (collections.defaultdict): The current latent weights assigned to the features.

    Example:

        ::

            >>> from creme import facto

            >>> X_y = (
            ...     ({'user': 'Alice', 'item': 'Superman', 'time': .12}, 8),
            ...     ({'user': 'Alice', 'item': 'Terminator', 'time': .13}, 9),
            ...     ({'user': 'Alice', 'item': 'Star Wars', 'time': .14}, 8),
            ...     ({'user': 'Alice', 'item': 'Notting Hill', 'time': .15}, 2),
            ...     ({'user': 'Alice', 'item': 'Harry Potter ', 'time': .16}, 5),
            ...     ({'user': 'Bob', 'item': 'Superman', 'time': .13}, 8),
            ...     ({'user': 'Bob', 'item': 'Terminator', 'time': .12}, 9),
            ...     ({'user': 'Bob', 'item': 'Star Wars', 'time': .16}, 8),
            ...     ({'user': 'Bob', 'item': 'Notting Hill', 'time': .10}, 2)
            ... )

            >>> model = facto.HOFMRegressor(
            ...     degree=3,
            ...     n_factors=10,
            ...     intercept=5,
            ...     seed=42,
            ... )

            >>> for x, y in X_y:
            ...     _ = model.fit_one(x, y)

            >>> model.predict_one({'user': 'Bob', 'item': 'Harry Potter', 'time': .14})
            5.311745

    Note:
        - For more efficiency, FM models automatically one hot encode string values considering them as categorical variables.
        - For model stability and better accuracy, numerical features should often be transformed into categorical ones.

    References:
        1. `Rendle, S., 2010, December. Factorization machines. In 2010 IEEE International Conference on Data Mining (pp. 995-1000). IEEE. <https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf>`_

    """

    def __init__(self, degree=3, n_factors=10, weight_optimizer=None, latent_optimizer=None,
                 loss=None, sample_normalization=False, l1_weight=0., l2_weight=0., l1_latent=0.,
                 l2_latent=0., intercept=0., intercept_lr=.01, weight_initializer=None,
                 latent_initializer=None, clip_gradient=1e12, seed=None):
        super().__init__(
            degree=degree,
            n_factors=n_factors,
            weight_optimizer=weight_optimizer,
            latent_optimizer=latent_optimizer,
            loss=optim.losses.Squared() if loss is None else loss,
            sample_normalization=sample_normalization,
            l1_weight=l1_weight,
            l2_weight=l2_weight,
            l1_latent=l1_latent,
            l2_latent=l2_latent,
            intercept=intercept,
            intercept_lr=intercept_lr,
            weight_initializer=weight_initializer,
            latent_initializer=latent_initializer,
            clip_gradient=clip_gradient,
            seed=seed
        )

    def predict_one(self, x):
        x = self._ohe_cat_features(x)
        return self._raw_dot(x)


class HOFMClassifier(HOFM, base.BinaryClassifier):
    """Higher-Order Factorization Machines Classifier.

    Parameters:
        degree (int): Polynomial degree or model order.
        n_factors (int): Dimensionality of the factorization or number of latent factors.
        weight_optimizer (optim.Optimizer): The sequential optimizer used for updating the feature
            weights. Note that the intercept is handled separately.
        latent_optimizer (optim.Optimizer): The sequential optimizer used for updating the latent
            factors.
        loss (optim.Loss): The loss function to optimize for.
        sample_normalization (bool): Whether to divide each element of ``x`` by ``x`` L2-norm.
            Defaults to False.
        l1_weight (float): Amount of L1 regularization used to push weights towards 0.
        l2_weight (float): Amount of L2 regularization used to push weights towards 0.
        l1_latent (float): Amount of L1 regularization used to push latent weights towards 0.
        l2_latent (float): Amount of L2 regularization used to push latent weights towards 0.
        intercept (float or `stats.Univariate` instance): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        weight_initializer (optim.initializers.Initializer): Weights initialization scheme. Defaults
            to ``optim.initializers.Zeros()``.
        latent_initializer (optim.initializers.Initializer): Latent factors initialization scheme.
            Defaults to
            ``optim.initializers.Normal(mu=.0, sigma=.1, random_state=self.random_state)``.
        clip_gradient (float): Clips the absolute value of each gradient value.
        seed (int): Randomization seed used for reproducibility.

    Attributes:
        weights (collections.defaultdict): The current weights assigned to the features.
        latents (collections.defaultdict): The current latent weights assigned to the features.

    Example:

        ::

            >>> from creme import facto

            >>> X_y = (
            ...     ({'user': 'Alice', 'item': 'Superman', 'time': .12}, True),
            ...     ({'user': 'Alice', 'item': 'Terminator', 'time': .13}, True),
            ...     ({'user': 'Alice', 'item': 'Star Wars', 'time': .14}, True),
            ...     ({'user': 'Alice', 'item': 'Notting Hill', 'time': .15}, False),
            ...     ({'user': 'Alice', 'item': 'Harry Potter ', 'time': .16}, True),
            ...     ({'user': 'Bob', 'item': 'Superman', 'time': .13}, True),
            ...     ({'user': 'Bob', 'item': 'Terminator', 'time': .12}, True),
            ...     ({'user': 'Bob', 'item': 'Star Wars', 'time': .16}, True),
            ...     ({'user': 'Bob', 'item': 'Notting Hill', 'time': .10}, False)
            ... )

            >>> model = facto.HOFMClassifier(
            ...     degree=3,
            ...     n_factors=10,
            ...     intercept=.5,
            ...     seed=42,
            ... )

            >>> for x, y in X_y:
            ...     _ = model.fit_one(x, y)

            >>> model.predict_one({'user': 'Bob', 'item': 'Harry Potter', 'time': .14})
            True

    Note:
        - For more efficiency, FM models automatically one hot encode string values considering them as categorical variables.
        - For model stability and better accuracy, numerical features should often be transformed into categorical ones.

    References:
        1. `Rendle, S., 2010, December. Factorization machines. In 2010 IEEE International Conference on Data Mining (pp. 995-1000). IEEE. <https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf>`_

    """

    def __init__(self, degree=3, n_factors=10, weight_optimizer=None, latent_optimizer=None,
                 loss=None, sample_normalization=False, l1_weight=0., l2_weight=0., l1_latent=0.,
                 l2_latent=0., intercept=0., intercept_lr=.01, weight_initializer=None,
                 latent_initializer=None, clip_gradient=1e12, seed=None):
        super().__init__(
            degree=degree,
            n_factors=n_factors,
            weight_optimizer=weight_optimizer,
            latent_optimizer=latent_optimizer,
            loss=optim.losses.Log() if loss is None else loss,
            sample_normalization=sample_normalization,
            l1_weight=l1_weight,
            l2_weight=l2_weight,
            l1_latent=l1_latent,
            l2_latent=l2_latent,
            intercept=intercept,
            intercept_lr=intercept_lr,
            weight_initializer=weight_initializer,
            latent_initializer=latent_initializer,
            clip_gradient=clip_gradient,
            seed=seed
        )

    def predict_proba_one(self, x):
        x = self._ohe_cat_features(x)
        p = utils.math.sigmoid(self._raw_dot(x))  # Convert logit to probability
        return {False: 1. - p, True: p}



import abc
import typing

from creme import base


class Recommender(base.Regressor):
    """A recommender."""

    def fit_one(self, x, y):
        return self._fit_one(x['user'], x['item'], y)

    def predict_one(self, x):
        return self._predict_one(x['user'], x['item'])

    @abc.abstractmethod
    def _fit_one(self, user: typing.Union[str, int], item: typing.Union[str, int],
                 y: float) -> 'Recommender':
        """Fits a `user`-`item` pair and a real-valued target `y`.

        Parameters:
            user: A user ID.
            item: An item ID.
            y: A rating.

        """

    @abc.abstractmethod
    def _predict_one(self, user: typing.Union[str, int], item: typing.Union[str, int]) -> float:
        """Predicts the target value of a set of features `x`.

        Parameters:
            user: A user ID.
            item: An item ID.

        Returns:
            The predicted rating.

        """



import importlib
import inspect
import math

import numpy as np
import pytest

from creme import optim


@pytest.mark.parametrize('loss', [
    pytest.param(loss(), id=name)
    for name, loss in inspect.getmembers(
        importlib.import_module('creme.optim.losses'),
        lambda x: inspect.isclass(x) and not inspect.isabstract(x) and not issubclass(x, optim.losses.CrossEntropy)
    )
])
def test_numpy(loss):

    y_true = np.random.uniform(-10, 10, 30)
    y_pred = np.random.uniform(-10, 10, 30)

    for yt, yp, g in zip(y_true, y_pred, loss.gradient(y_true, y_pred)):
        assert math.isclose(loss.gradient(yt, yp), g, abs_tol=1e-9)



from .. import utils

from . import base


__all__ = ['Newton']


class Newton(base.Optimizer):
    """Online Newton Step (ONS) optimizer.

    This optimizer uses second-order information (i.e. the Hessian of the cost function) in
    addition to first-order information (i.e. the gradient of the cost function).

    Parameters:
        lr
        eps

    References:
        1. [Hazan, E., Agarwal, A. and Kale, S., 2007. Logarithmic regret algorithms for online convex optimization. Machine Learning, 69(2-3), pp.169-192](https://www.cs.princeton.edu/~ehazan/papers/log-journal.pdf)

    """

    def __init__(self, lr=.1, eps=1e-5):
        super().__init__(lr)
        self.eps = eps
        self.H_inv = {}

    def _update_after_pred(self, w, g):

        for i in g:
            if (i, i) not in self.H_inv:
                self.H_inv[i, i] = self.eps

        # Update the Hessian
        self.H = utils.math.sherman_morrison(A_inv=self.H_inv, u=g, v=g)

        # Calculate the update step
        step = utils.math.dotvecmat(x=g, A=self.H_inv)

        # Update the weights
        for i, s in step.items():
            w[i] -= self.learning_rate * s

        return w



import collections

from . import base


__all__ = ['RMSProp']


class RMSProp(base.Optimizer):
    """RMSProp optimizer.

    Parameters:
        lr
        rho
        eps

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.RMSProp()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.872378

    References:
        1. [Divide the gradient by a running average of itsrecent magnitude](https://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)

    """

    def __init__(self, lr=.1, rho=.9, eps=1e-8):
        super().__init__(lr)
        self.rho = rho
        self.eps = eps
        self.g2 = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            self.g2[i] = self.rho * self.g2[i] + (1 - self.rho) * gi ** 2
            w[i] -= self.learning_rate / (self.g2[i] + self.eps) ** 0.5 * gi

        return w



from . import base


class Count(base.Univariate):
    """A simple counter.

    Attributes:
        n (int): The current number of observations.

    """

    def __init__(self):
        self.n = 0

    def update(self, x=None):
        self.n += 1
        return self

    def get(self):
        return self.n



import abc
import typing

from creme import base

from .predictor import Predictor


class Classifier(Predictor):
    """A classifier."""

    @property
    def _is_supervised(self):
        return True

    def predict_one(self, x: dict) -> typing.Optional[base.typing.ClfTarget]:
        """Predict the label of a set of features `x`.

        Parameters:
            x: A dictionary of features.

        Returns:
            The predicted label.

        """
        y_pred = self.predict_proba_one(x)
        if y_pred:
            return max(y_pred, key=y_pred.get)
        return None


class MultiClassifier(Classifier):
    """A multi-class classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: base.typing.ClfTarget) -> 'MultiClassifier':
        """Update the model with a set of features `x` and a label `y`.

        Parameters:
            x: A dictionary of features.
            y: A label.

        Returns:
            self

        """

    def predict_proba_one(self, x: dict) -> typing.Dict[base.typing.ClfTarget, float]:
        """Predict the probability of each label for a dictionary of features `x`.

        Parameters:
            x: A dictionary of features.

        Returns:
            A dictionary which associates a probability which each label.

        """
        raise NotImplementedError



class BinaryClassifier(Classifier):
    """A binary classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: bool) -> 'BinaryClassifier':
        """Update the model with a set of features `x` and a boolean value `y`.

        Parameters:
            x: A dictionary of features.
            y: A boolean value.

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_proba_one(self, x: dict) -> typing.Dict[bool, float]:
        """Predict the probability of both outcomes for a dictionary of features `x`.

        Parameters:
            x: A dictionary of features.

        Returns:
            A dictionary with the probabilities of `True` and `False`.

        """
        raise NotImplementedError



from ... import ensemble

from . import tree


__all__ = ['RandomForestClassifier']


class RandomForestClassifier(ensemble.BaggingClassifier):
    """Random forest classifier.

    This is a thin wrapper over `ensemble.BaggingClassifier` and `tree.DecisionTreeClassifier`.

    Parameters:
        n_trees: The number of trees in the forest.
        seed: Random number generator seed for reproducibility.
        tree_params: The parameters of the decision tree.

    Example:

        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import tree

        >>> X_y = datasets.Phishing()

        >>> model = tree.RandomForestClassifier(
        ...     n_trees=10,
        ...     seed=42,
        ...     # Tree parameters
        ...     patience=100,
        ...     confidence=1e-5,
        ...     criterion='gini'
        ... )

        >>> metric = metrics.LogLoss()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        LogLoss: 0.456533

    """

    def __init__(self, n_trees=10, seed: int = None, **tree_params: dict):
        super().__init__(
            model=tree.DecisionTreeClassifier(**tree_params),
            n_models=n_trees,
            seed=seed
        )

    @property
    def n_trees(self):
        return len(self)

    def _get_params(self):
        return {**super()._get_params(), **self.model._get_params()}

    def __str__(self):
        return 'RandomForestClassifier'



from . import base


__all__ = ['MSE']


class MSE(base.MeanMetric, base.RegressionMetric):
    """Mean squared error.

    Example:

        >>> from creme import metrics

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 8]

        >>> metric = metrics.MSE()

        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     print(metric.update(y_t, y_p).get())
        0.25
        0.25
        0.1666
        0.375

    """

    def _eval(self, y_true, y_pred):
        return (y_true - y_pred) ** 2



from creme import stream

from . import base


class Taxis(base.RemoteDataset):
    """Taxi ride durations in New York City.

    The goal is to predict the duration of taxi rides in New York City.

    References:
        1. [New York City Taxi Trip Duration competition on Kaggle](https://www.kaggle.com/c/nyc-taxi-trip-duration)

    """

    def __init__(self):
        super().__init__(
            n_samples=1_458_644,
            n_features=8,
            task=base.REG,
            url='https://maxhalford.github.io/files/datasets/nyc_taxis.zip',
            size=195271696,
            filename='train.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='trip_duration',
            converters={
                'passenger_count': int,
                'pickup_longitude': float,
                'pickup_latitude': float,
                'dropoff_longitude': float,
                'dropoff_latitude': float,
                'trip_duration': int
            },
            parse_dates={'pickup_datetime': '%Y-%m-%d %H:%M:%S'},
            drop=['dropoff_datetime', 'id']
        )



from .. import utils

from . import base


class Sum(base.Univariate):
    """Running sum.

    Attributes:
        sum (float): The running sum.

    Example:

        >>> from creme import stats

        >>> X = [-5, -3, -1, 1, 3, 5]
        >>> mean = stats.Sum()
        >>> for x in X:
        ...     print(mean.update(x).get())
        -5.0
        -8.0
        -9.0
        -8.0
        -5.0
        0.0

    """

    def __init__(self):
        self.sum = 0.

    def update(self, x):
        self.sum += x
        return self

    def get(self):
        return self.sum


class RollingSum(base.RollingUnivariate, utils.Window):
    """Running sum over a window.

    Parameters:
        window_size: Size of the rolling window.

    Attributes:
        sum (int): The running rolling sum.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 2, 1]
        >>> rolling_sum = stats.RollingSum(2)
        >>> for x in X:
        ...     print(rolling_sum.update(x).get())
        1
        -3
        -1
        1
        0
        3

    """

    def __init__(self, window_size: int):
        super().__init__(size=window_size)
        self.sum = 0

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        if len(self) == self.size:
            self.sum -= self[0]
        self.sum += x
        self.append(x)
        return self

    def get(self):
        return self.sum



"""Linear models."""
from .alma import ALMAClassifier
from .glm import LinearRegression
from .glm import LogisticRegression
from .pa import PAClassifier
from .pa import PARegressor
from .softmax import SoftmaxRegression


__all__ = [
    'ALMAClassifier',
    'LinearRegression',
    'LogisticRegression',
    'PAClassifier',
    'PARegressor',
    'SoftmaxRegression'
]



from . import base


class EWMean(base.Univariate):
    """Exponentially weighted mean.

    Parameters:
        alpha: The closer `alpha` is to 1 the more the statistic will adapt to recent values.

    Attributes:
        mean (float) : The running exponentially weighted mean.

    Example:

        >>> from creme import stats

        >>> X = [1, 3, 5, 4, 6, 8, 7, 9, 11]
        >>> ewm = stats.EWMean(alpha=0.5)
        >>> for x in X:
        ...     print(ewm.update(x).get())
        1
        2.0
        3.5
        3.75
        4.875
        6.4375
        6.71875
        7.859375
        9.4296875

    References:
        1. [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf)
        2. [Exponential Moving Average on Streaming Data](https://dev.to/nestedsoftware/exponential-moving-average-on-streaming-data-4hhl)

    """

    def __init__(self, alpha=.5):
        self.alpha = alpha
        self.mean = 0

    @property
    def name(self):
        return f'ewm_{self.alpha}'

    def update(self, x):
        self.mean = self.alpha * x + (1. - self.alpha) * self.mean if self.mean else x
        return self

    def get(self):
        return self.mean



import collections

from .. import utils

from . import base
from . import accuracy
from . import fbeta
from . import precision
from . import recall


class ClassificationReport(base.MultiClassMetric):
    """A report for monitoring a classifier.

    This class maintains a set of metrics and updates each of them every time `update` is called.
    You can print this class at any time during a model's lifetime to get a tabular visualization
    of various metrics.

    Parameters:
        decimals: The number of decimals to display in each cell.

    Example:

        >>> from creme import metrics

        >>> y_true = ['pear', 'apple', 'banana', 'banana', 'banana']
        >>> y_pred = ['apple', 'pear', 'banana', 'banana', 'apple']

        >>> report = metrics.ClassificationReport()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     report = report.update(yt, yp)

        >>> print(report)
                   Precision   Recall   F1      Support
        <BLANKLINE>
           apple       0.000    0.000   0.000         1
          banana       1.000    0.667   0.800         3
            pear       0.000    0.000   0.000         1
        <BLANKLINE>
           Macro       0.333    0.222   0.267
           Micro       0.400    0.400   0.400
        Weighted       0.600    0.400   0.480
        <BLANKLINE>
                         40.0% accuracy

    .. tip::
        You can wrap a `creme.metrics.ClassificationReport` with `creme.metrics.Rolling` in order
        to obtain a classification report over a window of observations. You can also wrap it with
        `creme.metrics.TimeRolling` to obtain a report over a period of time.

    """

    def __init__(self, decimals=3):
        self.decimals = decimals
        self._f1s = collections.defaultdict(fbeta.F1)
        self._macro_precision = precision.MacroPrecision()
        self._macro_recall = recall.MacroRecall()
        self._macro_f1 = fbeta.MacroF1()
        self._micro_precision = precision.MicroPrecision()
        self._micro_recall = recall.MicroRecall()
        self._micro_f1 = fbeta.MicroF1()
        self._weighted_precision = precision.WeightedPrecision()
        self._weighted_recall = recall.WeightedRecall()
        self._weighted_f1 = fbeta.WeightedF1()
        self._accuracy = accuracy.Accuracy()
        self._support = collections.Counter()

    def get(self):
        raise NotImplementedError

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred, sample_weight=1.):

        self._support[y_true] += 1

        # Update per class metrics
        for c in self._support:
            self._f1s[c].update(y_true == c, y_pred == c, sample_weight)

        # Update global metrics
        for m in [self._macro_precision, self._macro_recall, self._macro_f1,
                  self._micro_precision, self._micro_recall, self._micro_f1,
                  self._weighted_precision, self._weighted_recall, self._weighted_f1]:
            m.update(y_true, y_pred, sample_weight)

        self._accuracy.update(y_true, y_pred)

        return self

    def revert(self, y_true, y_pred, sample_weight=1.):

        self._support[y_true] -= 1

        # Revert per class metrics
        for c in self._support:
            self._f1s[c].revert(y_true == c, y_pred == c, sample_weight)

        # Revert global metrics
        for m in [self._macro_precision, self._macro_recall, self._macro_f1,
                  self._micro_precision, self._micro_recall, self._micro_f1,
                  self._weighted_precision, self._weighted_recall, self._weighted_f1]:
            m.revert(y_true, y_pred, sample_weight)
        self._accuracy.revert(y_true, y_pred)

        return self

    def __repr__(self):

        def fmt_float(x):
            return f'{x:.{self.decimals}f}'

        headers = ['', 'Precision', 'Recall', 'F1', 'Support']
        classes = sorted(self._support.keys())
        columns = [
            # Row names
            ['', *map(str, classes), '', 'Macro', 'Micro', 'Weighted'],
            # Precision values
            [
                '', *[fmt_float(self._f1s[c].precision.get()) for c in classes], '',
                *map(fmt_float, [
                    self._macro_precision.get(),
                    self._micro_precision.get(),
                    self._weighted_precision.get()
                ])
            ],
            # Recall values
            [
                '', *[fmt_float(self._f1s[c].recall.get()) for c in classes], '',
                *map(fmt_float, [
                    self._macro_recall.get(),
                    self._micro_recall.get(),
                    self._weighted_recall.get()
                ])
            ],
            # F1 values
            [
                '', *[fmt_float(self._f1s[c].get()) for c in classes], '',
                *map(fmt_float, [
                    self._macro_f1.get(),
                    self._micro_f1.get(),
                    self._weighted_f1.get()
                ])
            ],
            # Support
            ['', *[str(self._support[c]) for c in classes], *[''] * 4]
        ]

        # Build the table
        table = utils.pretty.print_table(headers, columns)

        # Write down the accuracy
        width = len(table.splitlines()[0])
        accuracy = f'{self._accuracy.get():.{self.decimals - 2}%}' + ' accuracy'
        table += '\n\n' + f'{accuracy:^{width}}'

        return table



import collections
import copy
import math

import numpy as np

from .. import base


__all__ = ['AdaBoostClassifier']


class BaseBoosting(base.Wrapper, base.Ensemble):

    def __init__(self, model, n_models=10, seed=None):
        super().__init__(copy.deepcopy(model) for _ in range(n_models))
        self.n_models = n_models
        self.model = model
        self.seed = seed
        self._rng = np.random.RandomState(seed)

    @property
    def _wrapped_model(self):
        return self.model


class AdaBoostClassifier(base.Classifier, BaseBoosting):
    """Boosting for classification

    For each incoming observation, each model's `fit_one` method is called `k` times where
    `k` is sampled from a Poisson distribution of parameter lambda. The lambda parameter is
    updated when the weaks learners fit successively the same observation.

    Parameters:
        model: The classifier to boost.
        n_models: The number of models in the ensemble.
        seed: Random number generator seed for reproducibility.

    Attributes:
        wrong_weight (collections.defaultdict): Number of times a model has made a mistake
            when making predictions.
        correct_weight (collections.defaultdict): Number of times a model has predicted the right
            label when making predictions.

    Example:

        In the following example three tree classifiers are boosted together. The performance is
        slightly better than when using a single tree.

        >>> from creme import datasets
        >>> from creme import ensemble
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import tree

        >>> X_y = datasets.Phishing()

        >>> metric = metrics.LogLoss()

        >>> model = ensemble.AdaBoostClassifier(
        ...     model=(
        ...         tree.DecisionTreeClassifier(
        ...             criterion='gini',
        ...             confidence=1e-5,
        ...             patience=2000
        ...         )
        ...     ),
        ...     n_models=5,
        ...     seed=42
        ... )

        >>> model_selection.progressive_val_score(X_y, model, metric)
        LogLoss: 0.741097

        >>> print(model)
        AdaBoostClassifier(DecisionTreeClassifier)

    References:
        1. [Oza, N.C., 2005, October. Online bagging and boosting. In 2005 IEEE international conference on systems, man and cybernetics (Vol. 3, pp. 2340-2345). Ieee.](https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf)

    """

    def __init__(self, model: base.Classifier, n_models=10, seed: int = None):
        super().__init__(model, n_models, seed)
        self.wrong_weight = collections.defaultdict(int)
        self.correct_weight = collections.defaultdict(int)

    def fit_one(self, x, y):
        lambda_poisson = 1

        for i, model in enumerate(self):
            for _ in range(self._rng.poisson(lambda_poisson)):
                model.fit_one(x, y)

            if model.predict_one(x) == y:
                self.correct_weight[i] += lambda_poisson
                lambda_poisson *= (
                    (self.correct_weight[i] + self.wrong_weight[i]) / (2 * self.correct_weight[i])
                )

            else:
                self.wrong_weight[i] += lambda_poisson
                lambda_poisson *= (
                    (self.correct_weight[i] + self.wrong_weight[i]) / (2 * self.wrong_weight[i])
                )
        return self

    def predict_proba_one(self, x):
        model_weights = collections.defaultdict(int)
        predictions = {}

        for i, model in enumerate(self):
            epsilon = self.correct_weight[i] + 1e-16
            epsilon /= (self.wrong_weight[i] + 1e-16)
            weight = math.log(epsilon)
            model_weights[i] += weight
            predictions[i] = model.predict_proba_one(x)

        y_pred = predictions[max(model_weights, key=model_weights.get)]
        total = sum(y_pred.values())
        return {label: proba / total for label, proba in y_pred.items()}



"""Running statistics"""
from .auto_corr import AutoCorr
from .base import Bivariate
from .base import Univariate
from .count import Count
from .cov import Cov
from .cov import RollingCov
from .entropy import Entropy
from .ewmean import EWMean
from .ewvar import EWVar
from .iqr import IQR
from .iqr import RollingIQR
from .link import Link
from .kurtosis import Kurtosis
from .maximum import Max
from .maximum import RollingMax
from .maximum import AbsMax
from .maximum import RollingAbsMax
from .mean import BayesianMean
from .mean import Mean
from .mean import RollingMean
from .minimum import Min
from .minimum import RollingMin
from .mode import Mode
from .mode import RollingMode
from .n_unique import NUnique
from .pearson import PearsonCorr
from .pearson import RollingPearsonCorr
from .ptp import PeakToPeak
from .ptp import RollingPeakToPeak
from .quantile import Quantile
from .quantile import RollingQuantile
from .shift import Shift
from .sem import SEM
from .sem import RollingSEM
from .skew import Skew
from .summing import Sum
from .summing import RollingSum
from .var import Var
from .var import RollingVar


__all__ = [
    'AbsMax',
    'AutoCorr',
    'BayesianMean',
    'Bivariate',
    'Count',
    'Cov',
    'Entropy',
    'EWMean',
    'EWVar',
    'IQR',
    'Kurtosis',
    'Link',
    'Max',
    'Mean',
    'Min',
    'Mode',
    'NUnique',
    'PeakToPeak',
    'PearsonCorr',
    'Quantile',
    'RollingAbsMax',
    'RollingCov',
    'RollingIQR',
    'RollingMax',
    'RollingMean',
    'RollingMin',
    'RollingMode',
    'RollingPeakToPeak',
    'RollingPearsonCorr',
    'RollingQuantile',
    'RollingSEM',
    'RollingSum',
    'RollingVar',
    'SEM',
    'Shift',
    'Skew',
    'Sum',
    'Univariate',
    'Var'
]



import collections
import typing

from . import base
from . import schedulers


__all__ = ['AMSGrad']


class AMSGrad(base.Optimizer):
    """AMSGrad optimizer.

    Parameters:
        lr: The learning rate.
        beta_1
        beta_2
        eps
        correct_bias

    Attributes:
        m (collections.defaultdict)
        v (collections.defaultdict)
        v_hat (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.AMSGrad()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.865724

    References:
        1. [Reddi, S.J., Kale, S. and Kumar, S., 2019. On the convergence of adam and beyond. arXiv preprint arXiv:1904.09237](https://arxiv.org/pdf/1904.09237.pdf)

    """

    def __init__(self, lr: typing.Union[float, schedulers.Scheduler] = .1,
                 beta_1=.9, beta_2=.999, eps=1e-8, correct_bias=True):
        super().__init__(lr)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.correct_bias = correct_bias
        self.m = collections.defaultdict(float)
        self.v = collections.defaultdict(float)
        self.v_hat = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        lr = self.learning_rate

        if self.correct_bias:
            # Correct bias for `v`
            lr *= (1 - self.beta_2 ** (self.n_iterations + 1)) ** .5
            # Correct bias for `m`
            lr /= 1 - self.beta_1 ** (self.n_iterations + 1)

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2
            self.v_hat[i] = max(self.v_hat[i], self.v[i])

            w[i] -= lr * self.m[i] / (self.v_hat[i] ** 0.5 + self.eps)

        return w



import itertools
import typing

import numpy as np


__all__ = ['expand_param_grid']


def expand_param_grid(grid: dict) -> typing.Iterator[dict]:
    """Expands a grid of parameters.

    This method can be used to generate a list of model parametrizations from a dictionary where
    each parameter is associated with a list of possible parameters. In other words, it expands a
    grid of parameters.

    Typically, this method can be used to create copies of a given model with different parameter
    choices. The models can then be used as part of a model selection process, such as a
    `model_selection.SuccessiveHalvingClassifier` or a `ensemble.HedgeRegressor`.

    The syntax for the parameter grid is quite flexible. It allows nesting parameters and can
    therefore be used to generate parameters for a pipeline.

    Parameters:
        grid: The grid of parameters to expand. The provided dictionary can be nested. The only
            requirement is that the values at the leaves need to be lists.

    Example:

        As an initial example, we can expand a grid of parameters for a single model.

        >>> import pprint
        >>> from creme import model_selection
        >>> from creme import optim

        >>> grid = {'optimizer': [optim.SGD(.1), optim.SGD(.01), optim.SGD(.001)]}

        >>> for params in model_selection.expand_param_grid(grid):
        ...     pprint.pprint(params)
        {'optimizer': SGD({'lr': Constant({'learning_rate': 0.1}), 'n_iterations': 0})}
        {'optimizer': SGD({'lr': Constant({'learning_rate': 0.01}), 'n_iterations': 0})}
        {'optimizer': SGD({'lr': Constant({'learning_rate': 0.001}), 'n_iterations': 0})}

        You can parametrize a model with a given set of parameters with the `_set_params` method:

        >>> from creme import linear_model

        >>> model = linear_model.LinearRegression(optimizer=optim.SGD())
        >>> model = model._set_params(params)
        >>> model
        LinearRegression (
          optimizer=SGD (
            lr=Constant (
              learning_rate=0.001
            )
          )
          loss=Squared ()
          l2=0.
          intercept=0.
          intercept_lr=Constant (
            learning_rate=0.01
          )
          clip_gradient=1e+12
          initializer=Zeros ()
        )

        You can expand parameters for multiple choices like so:

        >>> grid = {
        ...     'optimizer': [
        ...         (optim.SGD, {'lr': [.1, .01, .001]}),
        ...         (optim.Adam, {'lr': [.1, .01, .01]})
        ...     ]
        ... }

        You may specify a grid of parameters for a pipeline via nesting:

        >>> grid = {
        ...     'BagOfWords': {
        ...         'strip_accents': [False, True]
        ...     },
        ...     'LinearRegression': {
        ...         'optimizer': [
        ...             (optim.SGD, {'lr': [.1, .01]}),
        ...             (optim.Adam, {'lr': [.1, .01]})
        ...         ]
        ...     }
        ... }

    """

    def expand_tuple(t):

        klass, params = t

        if not isinstance(klass, type):
            raise ValueError(f'Expected first element to be a class, got {klass}')

        if not isinstance(params, dict):
            raise ValueError(f'Expected second element to be a dict, got {params}')

        return (klass(**combo) for combo in expand_param_grid(params))

    def expand(k, v):

        if isinstance(v, tuple):
            return ((k, el) for el in expand_tuple(v))

        # Example:
        # k = 'lr'
        # v = [0.001, 0.01, 0.1]
        if isinstance(v, (list, set, np.ndarray)):

            combos = []

            for el in v:
                if isinstance(el, tuple):
                    for combo in expand_tuple(el):
                        combos.append((k, combo))
                else:
                    combos.append((k, el))

            return combos

        if isinstance(v, dict):

            # Example:
            # k = 'LinearRegression'
            # v = {
            #     'intercept_lr': [0.001],
            #     'l2': [1],
            #     'optimizer': {
            #         optim.Adam: {
            #             'beta_1': [0.1, 0.01, 0.001],
            #             'lr': [0.1, 0.01, 0.001]
            #         },
            #      }
            # }
            return ((k, el) for el in expand_param_grid(v))

        raise ValueError(f'unsupported type: {type(v)}')

    for key in grid:
        if not isinstance(key, str):
            raise ValueError(f'Expected a key of type str; got {key}')

    # Example:
    # grid = {
    #     'beta_1': [0.1, 0.01, 0.001],
    #     'lr': [0.1, 0.01, 0.001]
    # }
    return (
        dict(el) if isinstance(el[0], tuple) else el[0]
        for el in itertools.product(*(expand(k, v) for k, v in grid.items()))
    )



"""Sampling methods."""
from .hard_sampling import HardSamplingClassifier
from .hard_sampling import HardSamplingRegressor

from .random import RandomOverSampler
from .random import RandomUnderSampler
from .random import RandomSampler


__all__ = [
    'HardSamplingClassifier',
    'HardSamplingRegressor',
    'RandomOverSampler',
    'RandomUnderSampler',
    'RandomSampler'
]
